# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import warnings
import random
import itertools
import math
from openexp.backend import Backend, configurable
from openexp.color import Color
from PIL import Image as PILImage
from collections import OrderedDict
from openexp.canvas_elements import (
    Line,
    Rect,
    Polygon,
    Ellipse,
    Image,
    Gabor,
    NoisePatch,
    Circle,
    FixDot,
    ElementFactory,
    RichText,
    Arrow
)
from openexp._canvas._element.element import Element
from openexp._canvas._element.group import Group


class Canvas(Backend):
    """
    {% set arg_max_width = "The maximum width of the text in pixels, " +
    "before wrapping to a new line, or `None` to wrap at screen edge." %}
    
    {% set arg_bgmode = "Specifies whether the background is the average of " +
    "col1 and col2 ('avg', corresponding to a typical Gabor patch), or " + 
    "equal to col2 ('col2'), useful for blending into the background. Note: " +
    "this parameter is ignored by the psycho backend, which uses increasing " + 
    "transparency for the background." %}
    
    {% set arg_style = "Optional [style keywords](#style-keywords) that " + 
    "specify the style of the current drawing operation. This does not " + 
    "affect subsequent drawing operations." %}
    
    {% set arg_center = "A bool indicating whether the coordinates reflect " + 
    "the center (`True`) or top-left (`False`) of the text." %}


    The `Canvas` class is used to present visual stimuli. You generally create a
    `Canvas` object with the `Canvas()` factory function, as described in the section
    [Creating a Canvas](#creating-a-canvas).
    
    __Example__:
    
    ~~~ .python
    # Create and show a canvas with a central fixation dot
    my_canvas = Canvas()
    my_canvas.fixdot()
    my_canvas.show()
    ~~~
    
    __Example__:
    
    You can also add `Canvas` elements as objects. See also the section on [Naming,
    accessing, and modifying elements](#naming-accessing-and-modifying-elements).
    
    ~~~ .python
    # Create a canvas with a fixation dot and a rectangle
    my_canvas = Canvas()
    my_canvas['my_fixdot'] = FixDot()
    my_canvas.show()
    ~~~
    
    [TOC]
    
    ## Things to know
    
    ### Creating a Canvas
    
    You generally create a `Canvas` with the `Canvas()` factory function:
    
    ~~~ .python
    my_canvas = Canvas()
    ~~~
    
    Optionally, you can pass [Style keywords](#style-keywords) to `Canvas()` to set
    the default style:
    
    ~~~ .python
    my_canvas = Canvas(color='green')
    my_canvas.fixdot() # Will be green
    ~~~
    
    ### Style keywords
    
    All functions that accept `**style_args` take the following keyword arguments:
    
    - `color` specifies the foreground color. For valid color specifications, see
      [colors](#colors).
    - `background_color` specifies the background color. For valid color
      specifications, see [colors](#colors).
    - `fill` indicates whether rectangles, circles, polygons, and ellipses are
      filled (`True`), or drawn as an outline (`False`).
    - `penwidth` indicates a penwidth in pixels and should be `int` or `float`.
    - `html` indicates whether HTML-tags are interpreted, and should be `True` or
      `False`.
    - `font_family` is the name of a font family, such as 'sans'.
    - `font_size` is a font size in pixels.
    - `font_italic` indicates whether text should italics, and should be `True` or
      `False`.
    - `font_bold` indicates whether text should bold, and should be `True` or
      `False`.
    - `font_underline` indicates whether text should underlined, and should be
      `True` or `False`.
    
    ~~~ .python
    # Draw a green fixation dot
    my_canvas = Canvas()
    my_canvas.fixdot(color='green')
    my_canvas.show()
    ~~~
    
    Style keywords only affect the current drawing operation (except when passed to
    `Canvas()` while creating the `Canvas`). To change the style for all subsequent
    drawing operations, set style properties, such as `canvas.color`, directly:
    
    ~~~ .python
    # Draw a red cross with a 2px penwidth
    my_canvas = Canvas()
    my_canvas.color = 'red'
    my_canvas.penwidth = 2
    my_canvas.line(-10, -10, 10, 10)
    my_canvas.line(-10, 10, 10, -10)
    my_canvas.show()
    ~~~
    
    Or pass the style properties to `Canvas()`:
    
    ~~~ .python
    # Draw a red cross with a 2px penwidth
    my_canvas = Canvas(color='red', penwidth=2)
    my_canvas.line(-10, -10, 10, 10)
    my_canvas.line(-10, 10, 10, -10)
    my_canvas.show()
    ~~~
    
    ### Colors
    
    You can specify colors in the following ways. This includes CSS3-type color
    specifications, but also supports some extra specifications, such as CIE l* a*
    b* color space.
    
    __Version note:__ The `hsv`, `hsl`, and `lab` color spaces are new in v3.3.0.
    
    - __Color names:__ 'red', 'black', etc. A full list of valid color names can be
      found [here](http://www.w3.org/TR/SVG11/types.html#ColorKeywords).
    - __Seven-character hexadecimal strings:__ `#FF0000`, `#000000`, etc. Here,
      values range from `00` to `FF`, so that `#FF0000` is bright red.
    - __Four-character hexadecimal strings:__ `#F00`, `#000`, etc. Here, values
      range from '0' to 'F' so that `#F00` is bright red.
    - __RGB strings:__ `rgb(255,0,0)`, `rgb(0,0,0)`, etc. Here, values range from
      0 to 255 so that `rgb(255,0,0)` is bright red.
    - __RGB percentage strings:__ `rgb(100%,0%,0%)`, `rgb(0%,0%,0%)`, etc. Here,
      values range from 0% to 100% so that `rgb(100%,0%,0%)` is bright red.
    - __RGB tuples:__ `(255, 0, 0)`, `(0, 0 ,0)`, etc. Here, values range from `0`
      to `255` so that `(255,0,0)' is bright red.
    - __HSV strings:__ `hsv(120, 100%, 100%)`. In the [HSV](https://en.wikipedia.org/
      wiki/HSL_and_HSV) color space, the hue parameter is an angle from 0 to 359,
      and the saturation and value parameters are percentages from 0% to 100%.
    - __HSL strings:__ `hsl(120, 100%, 50%)`. In the [HSL](https://en.wikipedia.org/
      wiki/HSL_and_HSV) color space, the hue parameter is an angle from 0 to 359,
      and the saturation and lightness parameters are percentages from 0% to 100%.
    - __LAB strings:__ `lab(53, -20, 0)`. In the [CIELAB](https://en.wikipedia.org/
      wiki/CIELAB_color_space) color space, the parameters reflect lightness (`l*`),
      green-red axis (`a*`, negative is green), and blue-yellow axis (`b*`, negative
      is blue). This uses the D65 white point and the sRGB transfer function, as
      implemented [here](https://www.psychopy.org/_modules/psychopy/tools/
      colorspacetools.html).
    - __Luminance values:__  `255`, `0`, etc. Here, values range from `0` to `255`
      so that `255` is white.
    
    ~~~ .python
    # Various ways to specify green
    my_canvas.fixdot(color='green')  # Dark green
    my_canvas.fixdot(color='#00ff00')
    my_canvas.fixdot(color='#0f0')
    my_canvas.fixdot(color='rgb(0, 255, 0)')
    my_canvas.fixdot(color='rgb(0%, 100%, 0%)')
    my_canvas.fixdot(color='hsl(100, 100%, 50%)')
    my_canvas.fixdot(color='hsv(0, 100%, 100%)')
    my_canvas.fixdot(color='lab(53, -20, 0)')  # Dark green
    my_canvas.fixdot(color=(0, 255, 0))  # Specify a luminance value (white)
    ~~~
    
    ### Naming, accessing, and modifying elements
    
    As of OpenSesame 3.2, the `Canvas` supports an object-based interface that allows
    you to name elements, and to access and modify elements individually, without
    having to redraw the entire `Canvas`.
    
    For example, the following will first add a red `Line` element to a `Canvas`
    and show it, then change the color of the line to green and show it again,
    and then finally delete the line and show the canvas again (which is now blank).
    The name of the element (`my_line`) is used to refer to the element for all the
    operations.
    
    ~~~ .python
    my_canvas = Canvas()
    my_canvas['my_line'] = Line(-100, -100, 100, 100, color='red')
    my_canvas.show()
    clock.sleep(1000)
    my_canvas['my_line'].color = 'green'
    my_canvas.show()
    clock.sleep(1000)
    del my_canvas['my_line']
    my_canvas.show()
    ~~~
    
    You can also add an element without explicitly providing a name for it. In that
    case, a name is generated automatically (e.g. `stim0`).
    
    ~~~ .python
    my_canvas = Canvas()
    my_canvas += FixDot()
    my_canvas.show()
    ~~~
    
    If you add a list of elements, they will be automatically grouped together, and
    you can refer to the entire group by name.
    
    ~~~ .python
    my_canvas = Canvas()
    my_canvas['my_cross'] = [    Line(-100, 0, 100, 0),    Line(0, -100, 0, 100)]
    my_canvas.show()
    ~~~
    
    To check whether a particular `x,y` coordinate falls within the bounding
    rectangle of an element, you can use `in`:
    
    ~~~ .python
    my_mouse = Mouse(visible=True)
    my_canvas = Canvas()
    my_canvas['rect'] = Rect(-100, -100, 200, 200)
    my_canvas.show()
    button, (x, y), time = my_mouse.get_click()
    if (x, y) in my_canvas['rect']:
        print('Clicked in rectangle')
    else:
        print('Clicked outside of rectangle')
    ~~~
    
    You can also get a list of the names of all elements that contain an `x,y`
    coordinate, using the `Canvas.elements_at()` function, documented below.
    """
    # The maximum duration that Canvas.show() should take, which is set to one
    # frame duration on a 60 Hz monitor. If Canvas.show() takes longer, a
    # warning given.
    MAX_SHOW_DT = 16

    def __init__(self, experiment, auto_prepare=True, **style_args):
        """Constructor to create a new `Canvas` object. You do not generally
        call this constructor directly, but use the `Canvas()` factory
        function, which is described here: [/python/common/]().

        Parameters
        ----------
        experiment : experiment
            The experiment object.
        auto_prepare : bool, optional
            Indicates whether the canvas should be automatically prepared after
            each drawing operation, so that [canvas.show] will be maximally
            efficient. If auto_prepare is turned off, drawing operations may be
            faster, but [canvas.show] will take longer, unless [canvas.prepare]
            is explicitly called in advance. This option exists mostly for
            historical purposes, because there are currently no backends for
            which it is necessary to disable auto prepare.
        **style_args : dict
            Optional [style keywords](#style-keywords), which will be used as
            the default for all drawing operations on this `Canvas`.


        Examples
        --------
        >>> my_canvas = Canvas()
        >>> my_canvas.fixdot()
        >>> my_canvas.show()
        """
        self.experiment = experiment
        self._width = self.experiment.var.width
        self._height = self.experiment.var.height
        self.auto_prepare = auto_prepare
        Backend.__init__(self, configurables={
            'color': None,
            'background_color': None,
            'fill': self.assert_bool,
            'penwidth': self.assert_numeric,
            'html': self.assert_bool,
            'font_family': self.assert_string,
            'font_size': self.assert_numeric,
            'font_italic': self.assert_bool,
            'font_bold': self.assert_bool,
            'font_underline': self.assert_bool,
        }, **style_args)
        self._elements = OrderedDict()
        self._stimnr = 0

    def __enter__(self):
        """The context manager provides an elegant way to disable auto
        preparation. However, it is hidden for now because none of the backends
        require it.
        """
        self._original_auto_prepare = self.auto_prepare
        self.auto_prepare = False
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """The context manager provides an elegant way to disable auto
        preparation. However, it is hidden for now because none of the backends
        require it.
        """
        self.auto_prepare = self._original_auto_prepare
        self.prepare()

    def __setitem__(self, key, value):
        """Sets a named element. If a list of elements is provided, a Group is
        automatically created.
        """
        if key in self._elements:
            raise ValueError('An element with the name "%s" already exists' %
                             key)
        if isinstance(value, list):
            value = Group(self, [e.construct(self)
                                 if isinstance(e, ElementFactory) else e for e in value])
        if isinstance(value, ElementFactory):
            value = value.construct(self)
        if not isinstance(value, Element):
            raise TypeError('%s is not a canvas element but %s' %
                            (key, type(value)))
        self._elements[key] = value

    def __delitem__(self, key):
        """Deletes an element by name."""
        del self._elements[key]

    def __getitem__(self, key):
        """Retrieves an element by name."""
        return self._elements[key]

    def __len__(self):
        """Gets the number of elements, where a Group counts as 1."""
        return len(self._elements)

    def __contains__(self, key):
        """Checks whether a named element is in the Canvas."""
        return key in self._elements

    def __iter__(self):
        """Iterates over the elements."""
        for name, stim in self._elements.items():
            yield name, stim

    def __iadd__(self, element):
        """Adds an unnamed element."""
        while 'stim%d' % self._stimnr in self._elements:
            self._stimnr += 1
        self['stim%d' % self._stimnr] = element
        return self

    def _get_name_element(self, element):
        """Gets a name, element tuple for an element or element name (i.e. get
        both if you have only one).

        Parameters
        ----------
        element : Element, str
            A SKETCHPAD element, or its name.

        Returns
        -------
        A (name, element) tuple.
        """
        if isinstance(element, str):
            return element, self._elements[element]
        for name, _element in self._elements.items():
            if element is _element:
                return name, element
        raise ValueError('"%s" not found in canvas"' % element)

    def elements_at(self, x, y):
        """*New in v3.2.0*

        Gets the names of elements that contain a
        particular `x, y`
        coordinate.

        Parameters
        ----------
        x : int, float
            An X coordinate.
        y : int, float
            A Y coordinate.

        Returns
        -------
        list
            A `list` of element names that contain the coordinate `x, y`.

        Examples
        --------
        >>> # Create and show a canvas with two partly overlapping rectangles
        >>> my_canvas = Canvas()
        >>> my_canvas['right_rect'] = Rect(x=-200, y=-100, w=200, h=200, color='red')
        >>> my_canvas['left_rect'] = Rect(x=-100, y=-100, w=200, h=200, color='green')
        >>> my_canvas.show()
        >>> # Collect a mouse click and print the names of the elements that
        >>> # contain the clicked point
        >>> my_mouse = Mouse(visible=True)
        >>> button, pos, time = my_mouse.get_click()
        >>> if pos is not None:
        >>>     x, y = pos
        >>>     print('Clicked on elements: %s' % my_canvas.elements_at(x, y))
        """
        elements = []
        for name, element in self:
            try:
                if (x, y) in element:
                    elements.append(name)
            except NotImplementedError:
                pass
        return elements

    def lower_to_bottom(self, element):
        """Lowers an element to the bottom, so that it is drawn first; that
        is, it becomes the background.

        Parameters
        ----------
        element : Element, str
            A SKETCHPAD element, or its name.
        """
        first_name, first_element = self._get_name_element(element)
        self._elements.pop(first_name)
        _elements = OrderedDict()
        _elements[first_name] = first_element
        for name, element in self._elements.items():
            _elements[name] = element
        self._elements = _elements

    def raise_to_top(self, element):
        """Raises an element to the top, so that it is drawn last; that is, it
        becomes the foreground.

        Parameters
        ----------
        element : Element, str
            A SKETCHPAD element, or its name.
        """
        last_name, last_element = self._get_name_element(element)
        self._elements.pop(last_name)
        _elements = OrderedDict()
        for name, element in self._elements.items():
            _elements[name] = element
        _elements[last_name] = last_element
        self._elements = _elements

    def add_element(self, element, name=None):
        """An alternative to the dict and += API. For internal use. {nodoc}"""
        if name is None:
            self.__iadd__(element)
            return
        self._elements[name] = element

    def rename_element(self, old_name, new_name):
        """Renames an element."""
        element = self._elements.pop(old_name)
        i = 1
        name = new_name
        while name in self._elements:
            name = '%s_%d' % (new_name, i)
            i += 1
        self._elements[name] = element

    def set_config(self, **cfg):

        # Convert color to backend specific colors
        if 'color' in cfg and not hasattr(cfg['color'], 'backend_color'):
            cfg['color'] = Color(self.experiment, cfg['color'])
        if 'background_color' in cfg \
                and not hasattr(cfg['background_color'], 'backend_color'):
            cfg['background_color'] = Color(self.experiment,
                                             cfg['background_color'])
        Backend.set_config(self, **cfg)

    def default_config(self):

        return {
            'penwidth': 1,
            'fill': False,
            'html': True,
            'color': self.experiment.var.foreground,
            'background_color': self.experiment.var.background,
            'font_size': self.experiment.var.font_size,
            'font_family': self.experiment.var.font_family,
            'font_bold': self.experiment.var.font_bold == 'yes',
            'font_italic': self.experiment.var.font_italic == 'yes',
            'font_underline': self.experiment.var.font_underline == 'yes',
        }

    @property
    def size(self):
        """The size of the canvas as a `(width, height)` tuple. This is a
        read-only property.
        """
        return self._width, self._height

    @property
    def width(self):
        """The width of the canvas. This is a read-only property."""
        return self._width

    @property
    def height(self):
        """The height of the canvas. This is a read-only property."""
        return self._height

    @property
    def left(self):
        """The x coordinate of the left edge of the display. This is a read-
        only property.
        """
        return self._left

    @property
    def right(self):
        """The x coordinate of the right edge of the display. This is a read-
        only property.
        """
        return self._right

    @property
    def top(self):
        """The y coordinate of the top edge of the display. This is a read-
        only property.
        """
        return self._top

    @property
    def bottom(self):
        """The y coordinate of the bottom edge of the display. This is a read-
        only property.
        """
        return self._bottom

    def copy(self, canvas):
        """Turns the current `Canvas` into a copy of the passed `Canvas`.
        
        __Warning:__

        Copying `Canvas` objects can result in unpredictable behavior. In many
        cases, a better solution is to recreate multiple `Canvas` objects from
        scratch, and/ or to use the element interface to update `Canvas`
        elements individually.

        Parameters
        ----------
        canvas : canvas
            The `Canvas` to copy.

        Examples
        --------
        >>> my_canvas = Canvas()
        >>> my_canvas.fixdot(x=100, color='green')
        >>> my_copied_canvas = Canvas()
        >>> my_copied_canvas.copy(my_canvas)
        >>> my_copied_canvas.fixdot(x=200, color="blue")
        >>> my_copied_canvas.show()
        """
        self._elements = OrderedDict([
            (name, element.copy(self))
            for name, element in canvas._elements.items()
        ])
        self.set_config(**canvas.get_config())

    def prepare(self):
        """Finishes pending canvas operations (if any), so that a subsequent
        call to [canvas.show] is extra fast. It's only necessary to call this
        function if you have disabled `auto_prepare` when initializing the
        `Canvas`.
        """
        for name, element in self._elements.items():
            if element.visible:
                element.prepare()

    def show(self):
        """Shows, or 'flips', the canvas on the screen.

        Returns
        -------
        int, float
            A timestamp of the time at which the canvas actually appeared on
            the screen, or a best guess if precise temporal information is not
            available. For more information about timing, see </misc/timing>.
            Depending on the back-end the timestamp is an `int` or a `float`.

        Examples
        --------
        >>> my_canvas = Canvas()
        >>> my_canvas.fixdot()
        >>> t = my_canvas.show()
        >>> exp.set('time_fixdot', t)
        """
        raise NotImplementedError()

    @configurable
    def clear(self, **style_args):
        """Clears the canvas with the current background color. Note that it
        is generally faster to use a different canvas for each experimental
        display than to use a single canvas and repeatedly clear and redraw it.

        Parameters
        ----------
        **style_args : dict
            {{arg_style}}


        Examples
        --------
        >>> my_canvas = Canvas()
        >>> my_canvas.fixdot(color='green')
        >>> my_canvas.show()
        >>> clock.sleep(1000)
        >>> my_canvas.clear()
        >>> my_canvas.fixdot(color='red')
        >>> my_canvas.show()
        """
        raise NotImplementedError()

    def fixdot(self, x=None, y=None, style='default', **style_args):
        """Draws a fixation dot. The default style is medium-open.

        - 'large-
        filled' is a filled circle with a 16px radius.
        - 'medium-filled' is a
        filled circle with an 8px radius.
        - 'small-filled' is a filled circle
        with a 4px radius.
        - 'large-open' is a filled circle with a 16px radius
        and a 2px hole.
        - 'medium-open' is a filled circle with an 8px radius
        and a 2px hole.
        - 'small-open' is a filled circle with a 4px radius and
        a 2px hole.
        - 'large-cross' is 16px cross.
        - 'medium-cross' is an 8px
        cross.
        - 'small-cross' is a 4px cross.

        Parameters
        ----------
        x : int, NoneType, optional
            The X coordinate of the dot center, or None to draw a horizontally
            centered dot.
        y : int, NoneType, optional
            The Y coordinate of the dot center, or None to draw a vertically
            centered dot.
        style : str, unicode, optional
            The fixation-dot style. One of: default, large-filled,
            medium-
            filled, small-filled, large-open, medium-open,
            small-open, large-
            cross, medium-cross, or small-cross.
            default equals medium-open.
        **style_args : dict
            {{arg_style}}


        Examples
        --------
        >>> my_canvas = Canvas()
        >>> # Function interface
        >>> my_canvas.fixdot()
        >>> # Element interface
        >>> my_canvas['my_fixdot'] = FixDot()
        """
        self += FixDot(x=x, y=y, style=style, **style_args)
        return 'stim%d' % self._stimnr

    def circle(self, x, y, r, **style_args):
        """Draws a circle.

        Parameters
        ----------
        x : int
            The center X coordinate of the circle.
        y : int
            The center Y coordinate of the circle.
        r : int
            The radius of the circle.
        **style_args : dict
            {{arg_style}}


        Examples
        --------
        >>> my_canvas = Canvas()
        >>> # Function interface
        >>> my_canvas.circle(100, 100, 50, fill=True, color='red')
        >>> # Element interface
        >>> my_canvas['my_circle'] = Circle(100, 100, 50, fill=True, color='red')
        """
        self += Circle(x, y, r, **style_args)
        return 'stim%d' % self._stimnr

    def line(self, sx, sy, ex, ey, **style_args):
        """Draws a line.

        Parameters
        ----------
        sx : int
            The left X coordinate.
        sy : int
            The top Y coordinate.
        ex : int
            The right X coordinate.
        ey : int
            The bottom Y coordinate.
        **style_args : dict
            {{arg_style}}
        """
        self += Line(sx, sy, ex, ey, **style_args)
        return 'stim%d' % self._stimnr

    def arrow(self, sx, sy, ex, ey, body_length=0.8, body_width=.5,
              head_width=30, **style_args):
        """Draws an arrow. An arrow is a polygon consisting of 7 vertices,
        with an arrowhead pointing at (ex, ey).

        Parameters
        ----------
        sx : int
            The X coordinate of the arrow's base.
        sy : int
            The Y coordinate of the arrow's base.
        ex : int
            The X coordinate of the arrow's tip.
        ey : int
            The Y coordinate of the arrow's tip..
        body_length : float, optional
            Proportional length of the arrow body relative to the full arrow
            [0-1].
        body_width : float, optional
            Proportional width (thickness) of the arrow body relative to the
            full arrow [0-1].
        head_width : float, optional
            Width (thickness) of the arrow head in pixels.
        """
        self += Arrow(sx, sy, ex, ey, body_width=body_width,
                      body_length=body_length, head_width=head_width, **style_args)
        return 'stim%d' % self._stimnr

    def rect(self, x, y, w, h, **style_args):
        """Draws a rectangle.

        Parameters
        ----------
        x : int
            The left X coordinate.
        y : int
            The top Y coordinate.
        w : int
            The width.
        h : int
            The height.
        **style_args : dict
            {{arg_style}}


        Examples
        --------
        >>> my_canvas = Canvas()
        >>> # Function interface
        >>> my_canvas.rect(-10, -10, 20, 20, fill=True)
        >>> # Element interface
        >>> my_canvas['my_rect'] = Rect(-10, -10, 20, 20, fill=True)
        """
        self += Rect(x, y, w, h, **style_args)
        return 'stim%d' % self._stimnr

    def ellipse(self, x, y, w, h, **style_args):
        """Draws an ellipse.

        Parameters
        ----------
        x : int
            The left X coordinate.
        y : int
            The top Y coordinate.
        w : int
            The width.
        h : int
            The height.
        **style_args : dict
            {{arg_style}}


        Examples
        --------
        >>> my_canvas = Canvas()
        >>> # Function interface
        >>> my_canvas.ellipse(-10, -10, 20, 20, fill=True)
        >>> # Element interface
        >>> my_canvas['my_ellipse'] = Ellipse(-10, -10, 20, 20, fill=True)
        """
        self += Ellipse(x, y, w, h, **style_args)
        return 'stim%d' % self._stimnr

    def polygon(self, vertices, **style_args):
        """Draws a polygon that defined by a list of vertices. I.e. a shape of
        points connected by lines.

        Parameters
        ----------
        vertices : list
            A list of tuples, where each tuple corresponds to a vertex. For
            example, [(100,100), (200,100), (100,200)] will draw a triangle.
        **style_args : dict
            {{arg_style}}


        Examples
        --------
        >>> my_canvas = Canvas()
        >>> n1 = 0,0
        >>> n2 = 100, 100
        >>> n3 = 0, 100
        >>> # Function interface
        >>> my_canvas.polygon([n1, n2, n3])
        >>> # Element interface
        >>> my_canvas['my_polygon'] = Polygon([n1, n2, n3])
        """
        self += Polygon(vertices, **style_args)
        return 'stim%d' % self._stimnr

    def text_size(self, text, center=True, max_width=None, **style_args):
        """Determines the size of a text string in pixels.

        Parameters
        ----------
        text : str, unicode
            A string of text.
        center : bool, optional
            {{arg_center}}
        max_width : int, NoneType, optional
            {{arg_max_width}}
        **style_args : dict
            {{arg_style}}


        Returns
        -------
        tuple
            A (width, height) tuple containing the dimensions of the text
            string.

        Examples
        --------
        >>> my_canvas = Canvas()
        >>> w, h = my_canvas.text_size('Some text')
        """
        return RichText(
            text,
            center=center,
            max_width=max_width,
            font_size=self.font_size,
            font_family=self.font_family,
            font_bold=self.font_bold,
            font_italic=self.font_italic
        ).construct(self.__class__(self.experiment)).size

    def text(self, text, center=True, x=None, y=None, max_width=None,
             **style_args):
        """Draws text.

        Parameters
        ----------
        text : str, unicode
            A string of text. When using Python 2, this should be either
            `unicode` or a utf-8-encoded `str`. When using Python 3, this
            should be either `str` or a utf-8-encoded `bytes`.
        center : bool, optional
            {{arg_center}}
        x : int, NoneType, optional
            The X coordinate, or None to draw horizontally centered text.
        y : int, NoneType, optional
            The Y coordinate, or None to draw vertically centered text.
        max_width : int, NoneType, optional
            {{arg_max_width}}
        **style_args : dict
            {{arg_style}}


        Examples
        --------
        >>> my_canvas = Canvas()
        >>> # Function interface
        >>> my_canvas.text('Some text with <b>boldface</b> and <i>italics</i>')
        >>> # Element interface
        >>> my_canvas['my_text'] = Text('Some text with <b>boldface</b> and <i>italics</i>')
        """
        self += RichText(text, center=center, x=x, y=y, max_width=max_width,
                         **style_args)
        return 'stim%d' % self._stimnr

    def image(self, fname, center=True, x=None, y=None, scale=None,
              rotation=None):
        """Draws an image from file. This function does not look in the file
        pool, but takes an absolute path.

        Parameters
        ----------
        fname : str, unicode
            The filename of the image. When using Python 2, this should be
            either `unicode` or a utf-8-encoded `str`. When using Python 3,
            this should be either `str` or a utf-8-encoded `bytes`.
        center : bool, optional
            A bool indicating whether coordinates indicate the center (True) or
            top-left (False).
        x : int, NoneType, optional
            The X coordinate, or `None` to draw a horizontally centered image.
        y : int, NoneType, optional
            The Y coordinate, or `None` to draw a vertically centered image.
        scale : float, int, NoneType, optional
            The scaling factor of the image. `None` or 1 indicate the original
            size. 2.0 indicates a 200% zoom, etc.
        rotation : float, int, NoneType, optional
            The rotation of the image `None` or 0 indicate the original
            rotation. Positive values indicate a clockwise rotation in degrees.

        Examples
        --------
        >>> my_canvas = Canvas()
        >>> # Determine the absolute path:
        >>> path = exp.pool['image_in_pool.png']
        >>> # Function interface
        >>> my_canvas.image(path)
        >>> # Element interface
        >>> my_canvas['my_image'] = Image(path)
        """
        self += Image(
            fname, center=center, x=x, y=y, scale=scale, rotation=rotation
        )
        return 'stim%d' % self._stimnr

    def gabor(self, x, y, orient, freq, env='gaussian', size=96, stdev=12,
              phase=0, col1='white', col2='black', bgmode='avg'):
        """Draws a Gabor patch. Note: The exact rendering of the Gabor patch
        depends on the back-end.

        Parameters
        ----------
        x : int
            The center X coordinate.
        y : int
            The center Y coordinate.
        orient : float, int
            Orientation in degrees [0 .. 360]. This refers to a
            clockwise rotation from a vertical.
        freq : float, int
            Frequency in cycles/px of the sinusoid.
        env : str, unicode, optional
            The envelope that determines the shape of the patch. Can be
            "gaussian", "linear", "circular", or "rectangular".
        size : float, int, optional
            A size in pixels.
        stdev : float, int, optional
            Standard deviation in pixels of the gaussian. Only applicable to
            gaussian envelopes.
        phase : float, int, optional
            Phase of the sinusoid [0.0 .. 1.0].
        col1 : str, unicode, optional
            A color for the peaks.
        col2 : str, unicode, optional
            A color for the troughs. Note: The psycho back-end
            ignores this
            parameter and always uses the inverse of
            `col1` for the throughs.
        bgmode : str, unicode, optional
            {{arg_bgmode}}

        Examples
        --------
        >>> my_canvas = Canvas()
        >>> # Function interface
        >>> my_canvas.gabor(100, 100, 45, .05)
        >>> # Element interface
        >>> my_canvas['my_gabor'] = Gabor(100, 100, 45, .05)
        """
        self += Gabor(x, y, orient, freq, env=env, size=size,
                      stdev=stdev, phase=phase, col1=col1, col2=col2, bgmode=bgmode)
        return 'stim%d' % self._stimnr

    def noise_patch(self, x, y, env='gaussian', size=96, stdev=12,
                    col1='white', col2='black', bgmode='avg'):
        """Draws a patch of noise, with an envelope. The exact rendering of
        the noise patch depends on the back-end.

        Parameters
        ----------
        x : int
            The center X coordinate.
        y : int
            The center Y coordinate.
        env : str, unicode, optional
            The envelope that determines the shape of the patch. Can be
            "gaussian", "linear", "circular", or "rectangular".
        size : float, int, optional
            A size in pixels.
        stdev : float, int, optional
            Standard deviation in pixels of the gaussian. Only applicable to
            gaussian envelopes.
        col1 : str, unicode, optional
            The first color.
        col2 : str, unicode, optional
            The second color. Note: The psycho back-end ignores this
            parameter
            and always uses the inverse of `col1`.
        bgmode : str, unicode, optional
            {{arg_bgmode}}

        Examples
        --------
        >>> my_canvas = Canvas()
        >>> # Function interface
        >>> my_canvas.noise_patch(100, 100, env='circular')
        >>> # Element interface
        >>> my_canvas['my_noise_patch'] = NoisePatch(100, 100, env='circular')
        """
        self += NoisePatch(x, y, env=env, size=size, stdev=stdev,
                           col1=col1, col2=col2, bgmode=bgmode)
        return 'stim%d' % self._stimnr

    @staticmethod
    def init_display(experiment):
        """{nodoc}Initializes the display before the experiment begins.

        Parameters
        ----------
        experiment : experiment
            An experiment object.
        """
        raise NotImplementedError()

    @staticmethod
    def close_display(experiment):
        """Closes the display after the experiment is finished. {nodoc}

        Parameters
        ----------
        experiment : experiment
            An experiment object.
        """
        raise NotImplementedError()


# Translation mapping from envelope names
ENV_SYNONYMS = {
    "c": "c",
    "circular": "c",
    "round": "c",
    "g": "g",
    "gaussian": "g",
    "gauss": "g",
    "normal": "g",
    "r": "r",
    "rectangular": "r",
    "rectangle": "r",
    "g": "g",
    "rect": "g",
    "square": "g",
    None: "g",
    "l": "l",
    "linear": "l",
    "lin": "l",
    "ln": "l",
    "l": "l"
}

canvas_cache = {}


def _color(col):
    """Wrapper function for _gabor and _noise_patch to convert color names to
    PyGame color objects.

    Parameters
    ----------
    col
        A color specification.

    Returns
    -------
    A PyGame color object.
    """
    from openexp._color.legacy import Legacy
    return Legacy(None, col).backend_color


def _xyr(x, y, size, orient=0):
    """Gets the coordinates relative to the center and compensating for the
    rotation.
    """
    dx = x - 0.5 * size
    dy = y - 0.5 * size
    t = math.atan2(dy, dx) + orient
    r = math.sqrt(dx ** 2 + dy ** 2)
    return r * math.cos(t), r * math.sin(t), r


def _gabor(
        orient,
        freq,
        env="gaussian",
        size=96,
        stdev=12,
        phase=0,
        col1="white",
        col2="black",
        bgmode="avg"
):
    """Returns a pygame surface containing a Gabor patch. For arguments, see
    [canvas.gabor].
    """
    env = _match_env(env)
    # Generating a Gabor patch takes quite some time, so keep
    # a cache of previously generated Gabor patches to speed up
    # the process.
    global canvas_cache
    key = "gabor_%s_%s_%s_%s_%s_%s_%s_%s_%s" % (
        orient,
        freq,
        env,
        size,
        stdev,
        phase,
        col1,
        col2,
        bgmode
    )
    if key in canvas_cache:
        return canvas_cache[key]
    im = PILImage.new('RGB', (size, size))
    px = im.load()
    # Convert the orientation to radians
    orient = math.radians(-orient)
    col1 = _color(col1)
    col2 = _color(col2)
    # rx and ry reflect the real coordinates in the
    # target image
    for rx, ry in itertools.product(range(size), range(size)):
        ux, uy, r = _xyr(rx, ry, size, orient)
        # Get the amplitude without the envelope (0 .. 1)
        amp = 0.5 + 0.5 * math.cos(2.0 * math.pi * (ux * freq + phase))
        # The envelope adjustment
        if env == "g":
            f = math.exp(-0.5 * (ux / stdev) ** 2 - 0.5 * (uy / stdev) ** 2)
        elif env == "l":
            f = max(0, (0.5 * size - r) / (0.5 * size))
        elif env == "c":
            if (r > 0.5 * size):
                f = 0.0
            else:
                f = 1.0
        else:
            f = 1.0
        # Apply the envelope
        if bgmode == "avg":
            amp = amp * f + 0.5 * (1.0 - f)
        elif bgmode == "col2":
            amp = amp * f
        else:
            raise ValueError(
                "Invalid argument for bgmode: %s "
                "(should be one of 'avg','col2')"
                % bgmode
            )
        r = col1.r * amp + col2.r * (1.0 - amp)
        g = col1.g * amp + col2.g * (1.0 - amp)
        b = col1.b * amp + col2.b * (1.0 - amp)
        px[rx, ry] = int(r), int(g), int(b)
    canvas_cache[key] = im
    return im


def _noise_patch(
        env="gaussian",
        size=96,
        stdev=12,
        col1="white",
        col2="black",
        bgmode="avg"
):
    """Returns a pygame surface containing a noise patch. For arguments, see
    [canvas.noise_patch].
    """
    env = _match_env(env)
    # Generating a noise patch takes quite some time, so keep
    # a cache of previously generated noise patches to speed up
    # the process.
    global canvas_cache
    key = "noise_%s_%s_%s_%s_%s_%s" % (env, size, stdev, col1, col2, bgmode)
    if key in canvas_cache:
        return canvas_cache[key]
    # Create a surface
    im = PILImage.new('RGB', (size, size))
    px = im.load()
    col1 = _color(col1)
    col2 = _color(col2)
    # rx and ry reflect the real coordinates in the
    # target image
    for rx, ry in itertools.product(range(size), range(size)):
        ux, uy, r = _xyr(rx, ry, size)
        # Get the amplitude without the envelope (0 .. 1)
        amp = random.random()
        # The envelope adjustment
        if env == "g":
            f = math.exp(-0.5 * (ux / stdev) ** 2 - 0.5 * (uy / stdev) ** 2)
        elif env == "l":
            f = max(0, (0.5 * size - r) / (0.5 * size))
        elif env == "c":
            if (r > 0.5 * size):
                f = 0.0
            else:
                f = 1.0
        else:
            f = 1.0
        # Apply the envelope
        if bgmode == "avg":
            amp = amp * f + 0.5 * (1.0 - f)
        elif bgmode == "col2":
            amp = amp * f
        else:
            raise ValueError(
                "Invalid argument for bgmode: %s "
                "(should be one of 'avg','col2')"
                % bgmode
            )
        r = col1.r * amp + col2.r * (1.0 - amp)
        g = col1.g * amp + col2.g * (1.0 - amp)
        b = col1.b * amp + col2.b * (1.0 - amp)
        px[rx, ry] = int(r), int(g), int(b)
    canvas_cache[key] = im
    return im


def _match_env(env):
    """Translation between various envelope names.

    Parameters
    ----------
    env : str, unicode
        An envelope name.

    Returns
    -------
    unicode
        A standard envelope name ("c", "g", "r" or "l")
    """
    if env not in ENV_SYNONYMS:
        raise ValueError("'%s' is not a valid envelope" % env)
    return ENV_SYNONYMS[env]


# Non PEP-8 alias for backwards compatibility
canvas = Canvas



"""A generic form plug-in"""

category = "Form"
help_url = "manual/forms/custom"
icon = "os-form_base"
controls = [
    {
        "label": "Edit the script to modify the form",
        "type": "text"
    }
]
priority = 90



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
import pygame
import platform
from collections import OrderedDict
from openexp.backend import configurable
from openexp._canvas.canvas import Canvas
from openexp._coordinates.legacy import Legacy as LegacyCoordinates
from libopensesame.oslogging import oslogger

# PyGame 1.9.2 suffers from character encoding issues. To get around those, we
# pass file objects directly to PyGame, instead of paths. These file objects
# need to remain open in the case of fonts, otherwise the associated font
# object won't work anymore. So here we keep a list of all open file objects,
# which are closed when the display is closed. In addition, we keep track of
# font objects so that we don't repeatedly initialize the same font.
#
# A bug report regarding character encoding is here:
# - https://bitbucket.org/pygame/pygame/issues/196/\
#   font-file-name-containing-unicode-error
fileobjects = []
fonts = {}


class Legacy(Canvas, LegacyCoordinates):

    r"""This is a canvas backend built on top of PyGame. For function
    specifications and docstrings, see `openexp._canvas.canvas`.
    """
    # The settings variable is used by the GUI to provide a list of back-end
    # settings
    settings = {
        u"pygame_hwsurface": {
            u"name": u"Hardware surface",
            u"description": u"Create a hardware surface",
            u"default": u"yes"
        },
        u"pygame_doublebuf": {
            u"name": u"Double buffering",
            u"description": u"Use double buffering",
            u"default": u"yes"
        },
        u"pygame_window_frame": {
            u"name": u"Draw window frame",
            u"description": u"Draw a frame in window mode",
            u"default": u"yes",
        },
        u"pygame_window_pos": {
            u"name": u"Window position",
            u"description": u"Window position in window mode (format: 'x,y' or 'auto')",
            u"default": u"auto",
        }
    }

    def __init__(self, experiment, auto_prepare=True, **style_args):

        Canvas.__init__(
            self, experiment,
            auto_prepare=auto_prepare,
            **style_args
        )
        LegacyCoordinates.__init__(self)
        self.antialias = True
        self.surface = self.experiment.surface.copy()
        self.clear()
        if platform.system() == u'Darwin':
            self.show = self._show_macos

    def show(self):

        t0 = pygame.time.get_ticks()
        self.experiment.surface.blit(self.surface, (0, 0))
        self.experiment.last_shown_canvas = self.surface
        pygame.display.flip()
        t1 = pygame.time.get_ticks()
        if t1 - t0 > self.MAX_SHOW_DT:
            oslogger.warning('Canvas.show() took {0} ms'.format(t1 - t0))
        return t1

    def _show_macos(self):
        r"""On Mac OS, the display is sometimes not refreshed unless there is
        some interaction with the event loop. Therefore we implement this hack
        which is only used on Mac OS.
        """
        self.experiment.surface.blit(self.surface, (0, 0))
        self.experiment.last_shown_canvas = self.surface
        pygame.display.flip()
        pygame.event.pump()
        return pygame.time.get_ticks()

    def prepare(self):
        r"""Finishes pending canvas operations (if any), so that a subsequent
        call to [canvas.show] is extra fast. It's only necessary to call this
        function if you have disabled `auto_prepare` in [canvas.__init__].
        """
        self.surface.fill(self.background_color.backend_color)
        Canvas.prepare(self)

    def lower_to_bottom(self, element):

        Canvas.lower_to_bottom(self, element)
        self.redraw()

    def raise_to_top(self, element):

        Canvas.raise_to_top(self, element)
        self.redraw()

    def redraw(self):

        if not self.auto_prepare:
            return
        self.prepare()

    def set_config(self, **cfg):

        Canvas.set_config(self, **cfg)
        if hasattr(self, u'surface'):
            self.redraw()

    def copy(self, canvas):

        self.surface = canvas.surface.copy()
        Canvas.copy(self, canvas)

    @configurable
    def clear(self):

        self.surface.fill(self.background_color.backend_color)
        self._elements = OrderedDict()

    def _text_size(self, text):

        return self._font.size(text)

    def __delitem__(self, key):

        del self._elements[key]
        self.redraw()

    @staticmethod
    def init_display(experiment):

        # Initialize PyGame and set the Window icon
        pygame.init()
        surf = pygame.Surface((32, 32))
        surf.fill((255, 255, 255))
        pygame.draw.circle(surf, (0, 0, 255), (16, 16), 10, 4)
        pygame.display.set_icon(surf)
        # Determine the video mode
        mode = 0
        if experiment.var.get(
                u"pygame_hwsurface",
                u"yes",
                [u"yes", u"no"]
        ) == u"yes":
            mode = mode | pygame.HWSURFACE
            oslogger.info(u'enabling hardware surface')
        else:
            oslogger.info(u'not enabling hardware surface')
        if experiment.var.get(
                u"pygame_doublebuf",
                u"yes",
                [u"yes", u"no"]
        ) == u"yes":
            mode = mode | pygame.DOUBLEBUF
            oslogger.info(u'enabling double buffering')
        else:
            oslogger.info(u'not enabling double buffering')
        if pygame.display.mode_ok(experiment.resolution(), mode):
            oslogger.info(u'video mode ok')
        else:
            oslogger.warning(u'warning: video mode not ok')
        if experiment.var.fullscreen == u'yes':
            mode = mode | pygame.FULLSCREEN
        if experiment.var.get(
                u'pygame_window_frame',
                u'yes',
                [u'yes', u'no']
        ) == u'no':
            mode = mode | pygame.NOFRAME
        if experiment.var.get(u'pygame_window_pos', u'auto') != u'auto':
            os.environ[u'SDL_VIDEO_WINDOW_POS'] = experiment.var.get(
                u'pygame_window_pos'
            )
        # Create the window and the surface
        experiment.window = pygame.display.set_mode(
            experiment.resolution(),
            mode
        )
        pygame.display.set_caption(u'OpenSesame (legacy backend)')
        pygame.mouse.set_visible(False)
        experiment.surface = pygame.display.get_surface()
        # Disable mouse-motion events because they fill up the cue
        pygame.event.set_blocked(pygame.MOUSEMOTION)

    @staticmethod
    def close_display(experiment):

        while fileobjects:
            fileobjects.pop().close()
        while fonts:
            fonts.pop(list(fonts.keys())[0], None)
        pygame.display.quit()


# Non PEP-8 alias for backwards compatibility
legacy = Legacy



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.logger import Logger as LoggerRuntime
from libqtopensesame.items.qtplugin import QtPlugin
from libqtopensesame.widgets.logger_widget import LoggerWidget
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'logger', category=u'item')


class Logger(LoggerRuntime, QtPlugin):
    """GUI controls for the logger item."""
    
    description = _(u'Logs experimental data')
    help_url = u'manual/logging'
    lazy_init = True

    def __init__(self, name, experiment, string=None):
        LoggerRuntime.__init__(self, name, experiment, string)
        QtPlugin.__init__(self)

    def init_edit_widget(self):
        super().init_edit_widget(stretch=False)
        self.logger_widget = LoggerWidget(self)
        self.add_widget(self.logger_widget)
        self.auto_add_widget(self.logger_widget.ui.checkbox_auto_log,
                             'auto_log')
        # To avoid recursion, we connect to the parent function which doesn't
        # update the text edits
        self.logger_widget.update()
        self.logger_widget.ui.textedit_include.textChanged.connect(
            self._update_logvars)
        self.logger_widget.ui.textedit_exclude.textChanged.connect(
            self._update_logvars)
        
    def _extract_variables(self, script):
        variables = []
        for var in script.splitlines():
            var = var.strip()
            if var:
                variables.append(var)
        return variables
            
    def _update_logvars(self):
        self.logvars = self._extract_variables(
            self.logger_widget.ui.textedit_include.toPlainText())
        self.exclude_patterns = self._extract_variables(
            self.logger_widget.ui.textedit_exclude.toPlainText())
        self.update_script()

    def edit_widget(self):
        super().edit_widget()
        for item in self.experiment.items.values():
            if item.item_type == self.item_type and item is not self:
                self.extension_manager.fire(u'notify',
                    message=_(u'You have multiple unlinked loggers. This can lead to messy log files.'),
                    category=u'warning')
                break
        self.logger_widget.update()

    def apply_edit_changes(self):
        super().apply_edit_changes()
        self.logger_widget.update()


# Alias for backwards compatibility
logger = Logger


# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._canvas._ellipse.ellipse import Ellipse


class Circle(Ellipse):

    def __init__(self, canvas, x, y, r, **properties):

        properties = properties.copy()
        self._r = r
        properties.update({'x': x, 'y': y, 'w': 2 * r, 'h': 2 * r, 'r': r})
        self.prepare = self.circle_prepare(self.prepare)
        Ellipse.__init__(self, canvas, **properties)

    @property
    def r(self):
        return self._r

    @r.setter
    def r(self, val):
        self._r = val
        self.w = self.h = 2 * val

    @property
    def rect(self):
        return self.x-self.r, self.y-self.r, self.r*2, self.r*2

    def __contains__(self, xy):

        return ((xy[0]-self.x)**2+(xy[1]-self.y)**2)**0.5 <= self.r

    def circle_prepare(self, ellipse_prepare):
        r"""A decorator that converts the center coordinates used by the circle
        to the top-left coordinates used by the ellipse.
        """
        def inner():

            r = self._properties[u'w'] // 2
            self._properties[u'x'] -= r
            self._properties[u'y'] -= r
            ellipse_prepare()
            self._properties[u'x'] += r
            self._properties[u'y'] += r

        return inner

    @staticmethod
    def _getter(key, self):

        if key == u'r':
            return self._properties[u'w'] // 2
        return Ellipse._getter(key, self)

    @staticmethod
    def _setter(key, self, val):

        if key == u'r':
            self._properties[u'w'] = val * 2
            self._properties[u'h'] = val * 2
            return
        Ellipse._setter(key, self, val)



"""OpenSesame core extensions"""

from libopensesame.metadata import __version__

# The name of the packages to check for updates on conda and pip
packages = ['opensesame', 'opensesame-core', 'datamatrix', 'qdatamatrix',
            'python-pseudorandom', 'pseudorandom', 'python-qnotifications',
            'qnotifications', 'python-fileinspector', 'fileinspector',
            'pyqode.python', 'pyqode.core', 'pyqode3-python', 'pyqode3-core']



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtGui, QtWidgets
from libopensesame.exceptions import OSException
from libqtopensesame import sketchpad_elements
from libqtopensesame.widgets.sketchpad_element_button import \
    SketchpadElementButton
from libqtopensesame.widgets.base_widget import BaseWidget
from libqtopensesame.validators import CondValidator


class SketchpadWidget(BaseWidget):
    """The sketchpad controls. Most of the actual work is handled by
    libqtopensesame.misc.sketchpad_canvas.

    Parameters
    ----------
    sketchpad : sketchpad
        A sketchpad object.
    """
    
    def __init__(self, sketchpad):
        super().__init__(sketchpad.main_window, ui='widgets.sketchpad')
        self.sketchpad = sketchpad
        self.initialized = False
        self.margin = 50
        self.canvas = self.sketchpad.canvas
        self.arrow_cursor = QtGui.QCursor(
            self.theme.qpixmap('os-pointer', size=32), 8, 4)
        self.ui.graphics_view.setScene(self.canvas)
        self.ui.graphics_view.setMouseTracking(True)
        self.ui.button_pointer.clicked.connect(self.select_pointer_tool)
        self.ui.spinbox_zoom.valueChanged.connect(self.zoom)
        self.ui.spinbox_scale.valueChanged.connect(self.apply_scale)
        self.ui.spinbox_rotation.valueChanged.connect(self.apply_rotation)
        self.ui.spinbox_penwidth.valueChanged.connect(self.apply_penwidth)
        self.ui.edit_color.textEdited.connect(self.apply_color)
        self.ui.edit_show_if.verb = 'show'
        self.ui.edit_show_if.editingFinished.connect(self.apply_show_if)
        self.ui.edit_show_if.setValidator(
            CondValidator(self, default='always'))
        self.ui.edit_name.editingFinished.connect(self.apply_name)
        self.ui.spinbox_arrow_head_width.valueChanged.connect(
            self.apply_arrow_head_width)
        self.ui.spinbox_arrow_body_width.valueChanged.connect(
            self.apply_arrow_body_width)
        self.ui.spinbox_arrow_body_length.valueChanged.connect(
            self.apply_arrow_body_length)
        self.ui.checkbox_center.toggled.connect(self.apply_center)
        self.ui.checkbox_fill.toggled.connect(self.apply_fill)
        self.ui.checkbox_html.toggled.connect(self.apply_html)
        self.ui.widget_font.font_changed.connect(self.apply_font)
        self.ui.checkbox_grid.toggled.connect(self.apply_grid)
        self.ui.spinbox_grid.valueChanged.connect(self.apply_grid)
        self.ui.button_zoom_fit.clicked.connect(self.zoom_fit)
        self.ui.button_zoom_1.clicked.connect(self.zoom_1)
        # Set the minimum height of the settings widget to the height that it
        # has when all controls are visible. This prevents the display from
        # jumping.
        self.ui.widget_settings.adjustSize()
        self.ui.widget_settings.setMinimumHeight(
            self.ui.widget_settings.height())
        self.build_toolbar()
        self.selected_element_tool = None
        self.set_size()
        self.init_settings()
        self.select_pointer_tool()

    @property
    def elements(self):
        return self.sketchpad.elements

    def apply_grid(self):
        if not self.ui.checkbox_grid.isChecked():
            self.canvas.grid = 1
        else:
            self.canvas.grid = self.ui.spinbox_grid.value()
        self.draw(refresh=True)

    def apply_font(self, family, size, italic, bold):
        r"""Applies changes to the font settings.

        Parameters
        ----------
        family : QString
            Font family.
        size : int
            Font size
        italic : bool
            Font italic setting
        bold : bool
            Font bold setting
        """
        for element in self.sketchpad.selected_elements():
            element.set_property('font_family', str(family))
            if size >= 0:
                element.set_property('font_size', size)
            element.set_property('font_italic', italic, yes_no=True)
            element.set_property('font_bold', bold, yes_no=True)
        self.draw()

    def apply_html(self, html):
        for element in self.sketchpad.selected_elements():
            element.set_property('html', html, yes_no=True)
        self.draw()

    def apply_fill(self, fill):
        for element in self.sketchpad.selected_elements():
            element.set_property('fill', int(fill))
        self.draw()

    def apply_scale(self, scale):
        for element in self.sketchpad.selected_elements():
            element.set_property('scale', scale)
        self.draw()

    def apply_rotation(self, rotation):
        for element in self.sketchpad.selected_elements():
            element.set_property('rotation', rotation)
        self.draw()

    def apply_center(self, center):
        for element in self.sketchpad.selected_elements():
            element.set_property('center', int(center))
        self.draw()

    def apply_color(self, color):
        color = str(color)
        for element in self.sketchpad.selected_elements():
            element.set_property('color', color)
        self.draw()

    def apply_show_if(self):
        show_if = self.ui.edit_show_if.text()
        for element in self.sketchpad.selected_elements():
            element.set_property('show_if', show_if)
            self.ui.edit_show_if.setText(show_if)
        self.draw()

    def apply_name(self):
        name = self.ui.edit_name.text()
        for element in self.sketchpad.selected_elements():
            element.set_property('name', name)
        self.draw()

    def apply_penwidth(self, penwidth):
        for element in self.sketchpad.selected_elements():
            element.set_property('penwidth', penwidth)
        self.draw()

    def apply_arrow_body_width(self, arrow_body_width):
        for element in self.sketchpad.selected_elements():
            element.set_property('arrow_body_width', arrow_body_width)
        self.draw()

    def apply_arrow_head_width(self, arrow_head_width):
        for element in self.sketchpad.selected_elements():
            element.set_property('arrow_head_width', arrow_head_width)
        self.draw()

    def apply_arrow_body_length(self, arrow_body_length):
        for element in self.sketchpad.selected_elements():
            element.set_property('arrow_body_length', arrow_body_length)
        self.draw()

    def show_element_settings(self, element):
        r"""Shows all the settings that are required by an element, and updates
        these settings based on the element.

        Parameters
        ----------
        element
            A sketchpad element.
        """
        # Float/ int properties need to be checked before they are applied
        for prop, _type in [
                ('penwidth', int),
                ('arrow_head_width', float),
                ('arrow_body_width', float),
                ('arrow_body_length', float),
                ('scale', float),
                ('rotation', int),
                ('show_if', str),
                ('name', str),
                ('color', str),
                ('center', bool),
                ('fill', bool)
        ]:
            # Check if the selected element requires the property
            req_fnc = getattr(element, 'requires_%s' % prop)
            if not req_fnc():
                continue
            # Check if the property can be retrieved. If not, that means it's
            # variably defined, or it has an invalid type.
            try:
                val = element.get_property(prop, _type=_type)
            except OSException:
                continue
            # Adjust the widget
            if _type in (int, float):
                spinbox = getattr(self.ui, 'spinbox_%s' % prop)
                if val is None:
                    spinbox.setEnabled(False)
                    spinbox.setValue(0)
                elif isinstance(val, (int, float)):
                    spinbox.setEnabled(True)
                    if isinstance(spinbox, QtWidgets.QSpinBox):
                        val = int(val)
                    spinbox.setValue(val)
            elif _type == bool:
                checkbox = getattr(self.ui, 'checkbox_%s' % prop)
                if val is None:
                    checkbox.setEnabled(False)
                    checkbox.setChecked(False)
                else:
                    checkbox.setEnabled(True)
                    checkbox.setChecked(val)
            else:
                edit = getattr(self.ui, 'edit_%s' % prop)
                edit.setText(val)
        # Text needs special treatment
        if element.requires_text():
            family = element.get_property('font_family')
            size = element.get_property('font_size', _type=int)
            bold = element.get_property('font_bold', _type=bool)
            italic = element.get_property('font_italic', _type=bool)
            if None in (family, size, bold, italic):
                self.ui.widget_font.setEnabled(False)
            else:
                self.ui.widget_font.setEnabled(True)
                self.ui.widget_font.set_font(family=family, size=size,
                                             bold=bold, italic=italic)
            val = element.get_property('html', _type=bool)
            if val is None:
                self.ui.checkbox_html.setEnabled(False)
                self.ui.checkbox_html.setChecked(False)
            else:
                self.ui.checkbox_html.setEnabled(True)
                self.ui.checkbox_html.setChecked(val)
        self.show_element_tool_settings(element)

    def zoom_fit(self):
        r"""Sets the best-fitting zoom level."""
        w = self.sketchpad.var.width
        h = self.sketchpad.var.height
        l = -w/2
        t = -h/2
        self.ui.graphics_view.fitInView(l, t, w, h,
                                        mode=QtCore.Qt.KeepAspectRatio)
        zoom = self.ui.graphics_view.transform().m11()
        self.ui.spinbox_zoom.setValue(zoom)

    def zoom_1(self):
        r"""Sets the zoom level to 1 (no zoom)."""
        self.zoom(1)
        self.ui.spinbox_zoom.setValue(1)

    def zoom(self, value):
        r"""Sets the canvas zoom level.

        Parameters
        ----------
        value : float
            The zoom level, where 1 is normaal zoom.
        """
        t = QtGui.QTransform()
        t.scale(value, value)
        self.ui.graphics_view.setTransform(t, combine=False)

    def zoom_diff(self, diff, adjust=.0005):
        r"""Changes the zoom level by a given proportion. This function is used
        mostly for zoom-by-scroll.

        Parameters
        ----------
        diff : float
            The zoom difference.
        adjust : float, optional
            A value that is multiplied with `diff` to get a sensible zoom
            speed.
        """
        zoom = self.ui.spinbox_zoom.value() + .0005 * diff
        self.ui.spinbox_zoom.setValue(zoom)

    def initialize(self):
        r"""Initializes the sketchpad widget."""
        if self.initialized:
            return
        self.ui.edit_color.initialize(experiment=self.experiment, parent=self)
        self.ui.widget_font.initialize(experiment=self.experiment, parent=self)
        self.initialized = True

    def init_settings(self):
        r"""Fills the controls with the correct starting values."""
        if self.ui.edit_color.text() == '':
            self.ui.edit_color.setText(self.sketchpad.var.get('foreground'))
        if self.ui.edit_show_if.text() == '':
            self.ui.edit_show_if.setText('always')

    def build_toolbar(self):
        r"""Builds the element-tool toolbar."""
        self.element_buttons = [self.ui.button_pointer]
        for element in sketchpad_elements.elements[::-1]:
            b = SketchpadElementButton(self, element)
            self.ui.layout_tools.insertWidget(1, b)
            self.element_buttons.append(b)

    def select_pointer_tool(self):
        r"""Selects the pointer tool."""
        self.unselect_all_tools()
        self.ui.button_pointer.setChecked(True)
        self.ui.widget_settings_text.setVisible(False)
        self.ui.widget_settings_color.setVisible(False)
        self.ui.widget_settings_penwidth.setVisible(False)
        self.ui.widget_settings_arrow_head_width.setVisible(False)
        self.ui.widget_settings_arrow_body_width.setVisible(False)
        self.ui.widget_settings_arrow_body_length.setVisible(False)
        self.ui.widget_settings_scale.setVisible(False)
        self.ui.widget_settings_rotation.setVisible(False)
        self.ui.widget_settings_fill.setVisible(False)
        self.ui.widget_settings_center.setVisible(False)
        self.ui.widget_settings_show_if.setVisible(False)
        self.ui.widget_settings_name.setVisible(False)
        self.ui.graphics_view.setCursor(self.arrow_cursor)

    def select_element_tool(self, element):
        r"""Selects an element tool."""
        self.selected_element_tool = element
        self.show_element_tool_settings(element)
        cursor = element.cursor()
        if isinstance(cursor, tuple):
            pixmap, hotx, hoty = cursor
            cursor = QtGui.QCursor(
                self.theme.qpixmap(pixmap, size=32), hotx, hoty)
        self.ui.graphics_view.setCursor(cursor)

    def show_element_tool_settings(self, element):
        r"""Shows the settings that are applicable to a given element."""
        self.ui.widget_settings_text.setVisible(element.requires_text())
        self.ui.widget_settings_color.setVisible(element.requires_color())
        self.ui.widget_settings_penwidth.setVisible(
            element.requires_penwidth())
        self.ui.widget_settings_arrow_head_width.setVisible(
            element.requires_arrow_head_width())
        self.ui.widget_settings_arrow_body_width.setVisible(
            element.requires_arrow_body_width())
        self.ui.widget_settings_arrow_body_length.setVisible(
            element.requires_arrow_body_length())
        self.ui.widget_settings_scale.setVisible(element.requires_scale())
        self.ui.widget_settings_rotation.setVisible(
            element.requires_rotation())
        self.ui.widget_settings_fill.setVisible(element.requires_fill())
        self.ui.widget_settings_center.setVisible(element.requires_center())
        self.ui.widget_settings_show_if.setVisible(element.requires_show_if())
        self.ui.widget_settings_name.setVisible(
            element.requires_name() and
            len(self.sketchpad.selected_elements()) < 2
        )

    def unselect_all_tools(self):
        r"""Unselects all tools."""
        self.selected_element_tool = None
        for b in self.element_buttons:
            b.setChecked(False)

    def sizeHint(self):
        r"""Provides a size hint for the widget, so that it can become small.

        Returns
        -------
        QSize
            A size hint.
        """
        return QtCore.QSize(0, 0)

    def set_size(self):
        r"""Sets the size of the QGraphicsView to the experiment resolution."""
        w = self.sketchpad.var.width
        h = self.sketchpad.var.height
        self.ui.graphics_view.setSceneRect(-self.margin-w/2, -self.margin-h/2,
                                           w+2*self.margin, h+2*self.margin)

    def draw(self, refresh=False):
        r"""Clears and redraws the sketchpad canvas.

        Parameters
        ----------
        refresh : bool, optional
            Indicates whether the entire sketchpad should be refreshed, which
            is necessary to repaint the background.
        """
        def z_sort(element):
            if isinstance(element.z_index, (int, float)):
                return -element.z_index
            return 0

        if refresh or \
                self.canvas.background_color != self.sketchpad.var.background:
            self.canvas.invalidate()
        self.canvas.clear()
        self.elements.sort(key=z_sort)
        for element in self.elements:
            element.draw()
        for notification in self.canvas.notifications:
            self.extension_manager.fire('notify', message=notification,
                                        category='info')
        self.sketchpad.apply_edit_changes()

    def center(self):
        r"""Centers the canvas view."""
        self.ui.graphics_view.centerOn(0, 0)

    def set_cursor_pos(self, xy):
        r"""Updates the cursor-position indicator.

        Parameters
        ----------
        xy
            An (x,y) tuple with the cursor position.
        """
        self.ui.label_cursor_pos.setText('%d,%d' % xy)

    def current_color(self):
        return str(self.ui.edit_color.text())

    def current_penwidth(self):
        return self.ui.spinbox_penwidth.value()

    def current_scale(self):
        return self.ui.spinbox_scale.value()

    def current_rotation(self):
        return self.ui.spinbox_rotation.value()

    def current_arrow_body_width(self):
        return self.ui.spinbox_arrow_body_width.value()

    def current_arrow_body_length(self):
        return self.ui.spinbox_arrow_body_length.value()

    def current_arrow_head_width(self):
        return self.ui.spinbox_arrow_head_width.value()

    def current_fill(self):
        return int(self.ui.checkbox_fill.isChecked())

    def current_center(self):
        return int(self.ui.checkbox_center.isChecked())

    def current_show_if(self):
        return self.ui.edit_show_if.text()

    def current_font_family(self):
        return self.ui.widget_font.family

    def current_font_size(self):
        return self.ui.widget_font.size

    def current_font_bold(self):
        return 'yes' if self.ui.widget_font.bold else 'no'

    def current_font_italic(self):
        return 'yes' if self.ui.widget_font.italic else 'no'

    def current_html(self):
        return 'yes' if self.ui.checkbox_html.isChecked() else 'no'


# Alias for backwards compatibility
sketchpad_widget = SketchpadWidget



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
import re
import sys
from libopensesame.oslogging import oslogger
from libopensesame.item_stack import item_stack_singleton
from libopensesame.py3compat import *
import traceback
import time
import markdown


class AbortCoroutines(Exception):
    r"""A messaging Exception to indicate that coroutines should be aborted.
    That is, if a task raises an AbortCoroutines, then the currently running
    coroutines should abort.
    """
    
    pass


class OSException(Exception):
    """An OSException is raised when an error occurs that is not covered by
    any of the other Exception classes.
    """

    def __init__(self, msg):
        """Constructor.

        Parameters
        ----------
        msg : str, unicode
            An error message.
        """
        super().__init__(msg)
        self._msg = msg
        try:
            self.item, self.phase = item_stack_singleton[-1]
        except IndexError:
            # This may happen when the item stack is empty
            self.item = self.phase = 'unknown'
        self._read_more = f'''
<div id="more-information"
style="display:none;">{markdown.markdown(self.__doc__)}</div>

<a id="read-more" class="button"
onclick='document.getElementById("more-information").style.display = "block";document.getElementById("read-more").style.display = "none"'>
Learn more about this error</a>
'''

    def __str__(self):
        return f'''
{self.title()}\n\n{self._msg}

This error occurred in the {self.phase} phase of item {self.item}.
'''

    def markdown(self):
        return f'''
# {self.title()}

{self._msg}

This error occurred in the __{self.phase}__ phase of item 
<u><a href="opensesame://item.{self.item}.{self.phase}">
{self.item}</a></u>.

{self._read_more}
'''
    
    def title(self):
        return f'Error: {self.__class__.__name__}'


class BackendNotSupported(OSException):
    """This exception is raised when functionality is not supported backend.
    """
    pass

    
class InvalidKeyName(OSException):
    """An `InvalidKeyName` is raised when the name of a key has been
    incorrectly specified.
    """
    def __init__(self, key):
        super().__init__(f'Invalid key name: {key}')
        
        
class InvalidValue(OSException):
    """An `InvalidValue` is raised when an invalid value has been assigned to a
    variable. This can occur in many situations and generally indicates a
    mistake in the experiment. For example, you will get an `InvalidValue` if
    you specify a negative `sketchpad` duration.
    """
    pass


class InvalidColor(OSException):
    """An `InvalidColor` is raised when you have specified a color with an
    incorrect. This generally indicates a mistake in the experiment, such as a
    typo. For example, you will get an `InvalidColor` if you specify the color
    'bleu' (as opposed to 'blue'). For a list of valid color specifications,
    please visit see the documentation site.
    """
    def __init__(self, color):
        super().__init__(f'Invalid color specification: {color}')
        
        
class InvalidSketchpadElementScript(OSException):
    """An `InvalidSketchpadElementScript` is raised when there is a mistake in
    the script that defines a sketchpad element. This can happen when you
    manually modify a `sketchpad` script.
    """
    pass


class InvalidFormScript(OSException):
    """An `InvalidFormScript` is raised when there is a mistake in a form
    script. This can happen when you make a mistake while manually building a
    form using the `form_base` item, or while modifying the script of other
    form items.
    """
    pass


class InvalidOpenSesameScript(OSException):
    """An `InvalidOpenSesameScript` is raised when there is an error in the 
    OpenSesame script that defines the experiment and the items. This can 
    happen when you make a mistake while modifying the experiment script.
    """
    pass


class InvalidFormGeometry(OSException):
    """An `InvalidFormGeometry` is raised when a form has an impossible 
    geometry. This can happen for different reasons. For example, you will get
    an `InvalidFormGeometry` when a form is too small to fit
    all of its widgets.
    """
    pass


class UnsupportedImageFormat(OSException):
    """An `UnsupportedImageFormat` is raised when an image file does not have
    a format that is supported by OpenSesame. It can also occur when a file
    is not an image file at all. To solve this issue, make sure that the
    indicated file is indeed an image, and if necessary convert it to a
    standard format, such as `.png` or `.jpg`.
    """
    def __init__(self, path):
        super().__init__(f'{path} is not a supported image file')
        
        
class ImageDoesNotExist(OSException):
    """An `ImageDoesNotExist` is raised when you have specified an image file
    that does not exist. This commonly reflects a mistake in the experiment,
    such as a typo in a file name.
    """
    def __init__(self, path):
        super().__init__(f'Cannot find image file {path}')
        

class UnsupportedSoundFileFormat(OSException):
    """An `UnsupportedSoundFileFormat` is raised when a sound file does not
    have a format that is supported by OpenSesame. It can also occur when a
    file is not a sound image file at all. To solve this issue, make sure that
    the indicated file is indeed a sound file, and if necessary convert it to a
    standard format, such as `.wav` or `.mp3`.
    """
    def __init__(self, path):
        super().__init__(f'{path} is not a supported sound file')
        
        
class SoundFileDoesNotExist(OSException):
    """A `SoundFileDoesNotExist` is raised when you have specified a sound file
    that does not exist. This commonly reflects a mistake in the experiment,
    such as a typo in a file name.
    """
    def __init__(self, path):
        super().__init__(f'Cannot find sound file {path}')
        
        
class LoopSourceFileDoesNotExist(OSException):
    """A `LoopSourceFileDoesNotExist` is raised when you have specified a
    non-existent source file for a loop item. This commonly reflects  mistake
    in the experiment, such as a typo in a file name.
    """
    def __init__(self, path):
        super().__init__(f'Cannot find loop source file {path}')
        

class UnsupportedLoopSourceFile(OSException):
    """An `UnsupportedLoopSourceFile` is raised when you have specified a 
    source file for a loop item that is not a `.csv` or `.xlsx` file.
    """
    def __init__(self, path):
        super().__init__(f'{path} is not a supported loop source file')


class UserAborted(OSException):
    """This exception is raised when a user aborts an experiment."""
    def __str__(self):
        return self._msg
        

class UserKilled(UserAborted):
    """This exception is raised when a user kills an experiment."""
    pass

        
class MissingDependency(OSException):
    """A `MissingDependency` is raised when some functionality requires a
    package that is not available.
    """
    pass


class ItemDoesNotExist(OSException):
    """An `ItemDoesNotExist` is raised when the experiment refers to an item
    that does not exist. This can occur for different reasons.
    """
    def __init__(self, item_name):
        super().__init__(f'Item {item_name} does not exist')
        
        
class VariableDoesNotExist(OSException):
    """A `VariableDoesNotExist` is raised when the experiment refers to a
    variable that does not, or not yet, exist. This commonly reflects a mistake
    in the experiment, such as a typo in the name of a variable, or referring
    to a variable before it has been created. Note that variable names are case
    sensitive.
    """
    def __init__(self, var_name):
        super().__init__(f'Variable {var_name} does not exist')
        
        
class DeviceError(OSException):
    """A `DeviceError` is raised when an error occurs while connecting to, or
    interacting with, an external device. This can occur for different reasons.
    """
    pass


class PythonError(OSException):
    """A `PythonError` is raised when an error occurs during execution of 
    Python code, typically in an `inline_script` item.
    """
    def __init__(self, msg):
        super().__init__(msg)
        tb_lines = traceback.format_exc().splitlines()
        # Get the line number of most recent stack from the traceback. If no
        # traceback exists, fall back to line number 1
        tb = sys.exc_info()[2]
        if tb is None:
            self.line_nr = 1
        else:
            while tb.tb_next is not None:
                tb = tb.tb_next
            self.line_nr = tb.tb_lineno
        self._traceback = self.clean_traceback(tb_lines)

    def clean_traceback(self, tb_lines):
        # The __ignore_traceback__ comment serves as a marker to indicate which
        # parts of the error message should be hidden from the user because
        # they relate to the insides of OpenSesame, rather than to the user
        # error.
        print(tb_lines)
        for line_nr, tb_line in enumerate(tb_lines):
            if '# __ignore_traceback__' in tb_line:
                # If the traceback has ^^^^ indicators below the lines of code
                # to indicate the offending part of the line, then this 
                # indicator needs to be skipped too. The code below checks 
                # whether the string consists only of ^ characters.
                caret_line = tb_lines[line_nr + 1].strip()
                if caret_line.count('^') == len(caret_line):
                    line_nr += 1
                tb_lines = tb_lines[:1] + tb_lines[line_nr + 1:]
                break
        return '\n'.join(tb_lines).replace('<string>',
                                           f'<{self.item}.{self.phase}>')
        
    def __str__(self):
        return f'''
{self.title()}

{self._msg}

This error occurred on line {self.line_nr} in the {self.phase} phase of item {self.item}.

{self._traceback}
'''
        
    def markdown(self):
        
        return f'''
# {self.title()}

{self._msg}

This error occurred on __line {self.line_nr}__ in the
__{self.phase}__ phase of item
<u><a href="opensesame://item.{self.item}.{self.phase}.{self.line_nr - 1}">{self.item}</a></u>.

~~~ .traceback
{self._traceback}
~~~

{self._read_more}
'''


class PythonSyntaxError(PythonError):
    """A `PythonSyntaxError` is raised when a Python script, typically in an
    `inline_script` item, is not syntactically correct.
    """
    def __init__(self, msg, line_nr=1):
        super().__init__(msg)
        tb_lines = traceback.format_exc().splitlines()
        self.line_nr = line_nr
        self._traceback = self.clean_traceback(tb_lines)


class InvalidConditionalExpression(PythonSyntaxError):
    """An `InvalidConditionalExpression` is raised when a conditional 
    expression, such as a run-if, break-if, or show-if expression, is
    syntactically incorrect. This generally reflects a mistake in the 
    experiment, such as a typo in a conditional expression that renders it
    invalid.
    """
    def markdown(self):
        
        return f'''
# {self.title()}

{self._msg}

This error occurred in the
__{self.phase}__ phase of item
<u><a href="opensesame://item.{self.item}.{self.phase}.{self.line_nr - 1}">{self.item}</a></u>.

~~~ .traceback
{self._traceback}
~~~

{self._read_more}
'''


class ConditionalExpressionError(PythonError):
    """A `ConditionalExpressionError` is raised when an error occurs during the
    evaluation of a conditional expression, such as run-if, break-if, or
    show-if expression.
    """
    def markdown(self):
        
        return f'''
# {self.title()}

{self._msg}

This error occurred in the
__{self.phase}__ phase of item
<u><a href="opensesame://item.{self.item}.{self.phase}.{self.line_nr - 1}">{self.item}</a></u>.

~~~ .traceback
{self._traceback}
~~~

{self._read_more}
'''


class BaseFStringError:


    def __str__(self):
        return f'''
{self.title()}

{self._msg}

This error occurred in the {self.phase} phase of item {self.item}.

{self._traceback}
'''
        
    def markdown(self):
        return f'''
# {self.title()}

{self._msg}

This error occurred in the __{self.phase}__ phase of item
<u><a href="opensesame://item.{self.item}">{self.item}</a></u>.

~~~ .traceback
{self._traceback}
~~~

{self._read_more}
'''


class FStringError(BaseFStringError, PythonError):
    """An `FStringError` is raised when an error occurs during the evaluation
    of an f-string.
    """
    pass


class FStringSyntaxError(BaseFStringError, PythonSyntaxError):
    """An `FStringSyntaxError` is raised when text contains an f-string 
    expression that is not syntactically valid Python.
    """
    pass


class ExperimentProcessDied(OSException):
    """An `ExperimentProcessDied` is raised when the experiment process died.
    This is  generally the result of a bug in one of the underlying libraries 
    that causes Python to crash. This should not happen! If you experinence 
    this error often, please report it on the support forum.
    """
    pass


class UnexpectedError(OSException):
    """An `UnexpectedError` is raised when an error occurred that OpenSesame 
    does not recognize. This can happen when there is a bug in a plugin, one of
    the underlying Python libraries, or in OpenSesame itself. This should not
    happen! If you experinence this error often, please report it on the
    support forum.
    """
    def __init__(self, msg):
        super().__init__(msg)
        self._traceback = traceback.format_exc()
        
    def __str__(self):
        return f'{self.title()}\n\n{self._msg}\n\n{self._traceback}'
        
    def markdown(self):
        return f'# {self.title()}\n\n{self._msg}\n\n' \
               f'~~~ .traceback\n{self._traceback}\n~~~\n\n{self._read_more}'


class IncompatibilityError(OSException):
    """An `IncompatibilityError` is raised when the experiment uses 
    functionality that is not compatible with the current version of OpenSeame.
    """
    pass


# For backwards compatibility, we should also define the old Exception classes
osexception = OSException



"""Graphical variable inspector"""

icon = "os-inspector"
label = "Show variable inspector"
shortcut = "Ctrl+I"
checkable = True
menu = {
    "index": 8,
    "submenu": "View"
}
toolbar = {
    "index": 11
}
settings = {
    "variable_inspector_visible": False,
    "variable_inspector_focus_shortcut": "Ctrl+5"
}
tooltip = "Show variable inspector"



"""Updates the Quest test value based on a response"""

category = "Staircase"
controls = [
    {
        "type": "line_edit",
        "var": "response_var",
        "label": "Response variable",
        "info": "Responses should be coded as 0 (incorrect) or 1 (correct)"
    },
    {
        "type": "line_edit",
        "var": "quest_name",
        "label": "Name",
        "info": "Should match a name specified in a quest_staircase_init"
    }
]



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.widgets._label import Label


class TextInput(Label):

    r"""The text_input widget allows the participant to enter multi-character
    responses. (This widget has no relation to the text_input plug-in, which
    was created before forms where added to OpenSesame.)

    __Example (OpenSesame
    script):__

    ~~~
    widget 0 0 1 1 text_input var='response'
    return_accepts='yes'
    ~~~

    __Example (Python):__

    ~~~ .python
    form = Form()
    text_input = TextInput(var='response', return_accepts=True)
    form.set_widget(text_input, (0,0))
    form._exec()
    ~~~

    [TOC]
    """
    def __init__(
            self,
            form,
            text=u'',
            frame=True,
            center=False,
            stub=u'Type here ...',
            return_accepts=False,
            var=None,
            key_filter=None
    ):
        r"""Constructor to create a new `TextInput` object. You do not
        generally
        call this constructor directly, but use the `TextInput()`
        factory
        function, which is described here: [/python/common/]().

        Parameters
        ----------
        form : form
            The parent form.
        text : str, unicode, optional
            The text to start with.
        frame : bool, optional
            Indicates whether a frame should be drawn around the widget.
        center : bool, optional
            Indicates whether the text should be centered.
        stub : str, unicode, optional
            A text string that should be shown whenever the user has not
            entered any text.
        return_accepts : bool, optional
            Indicates whether a return press should accept and close the form.
        var : str, unicode, NoneType, optional
            The name of the experimental variable that should be used to log
            the widget status.
        key_filter : FunctionType, NoneType, optional
            A function that takes a key as a single argument and return True if
            the key should be accepted and False otherwise. This can also
            filter out keys such as return and backspace, but not Escape.
        """
        if isinstance(return_accepts, str):
            return_accepts = return_accepts == u'yes'
        Label.__init__(self, form, text, frame=frame, center=center)
        self.type = u'text_input'
        self.stub = safe_decode(stub)
        self.prompt = u'\u2038'
        self.html = False
        self.return_accepts = return_accepts
        self.var = var
        self.text = safe_decode(text)
        self.set_var(text)
        self.caret_pos = None
        self._key_filter = (
            lambda k: True) if key_filter is None else key_filter

    def _update(self):
        r"""Draws the widget."""
        if self.frame:
            if self.focus:
                self._update_frame(self.rect, style=u'active')
            else:
                self._update_frame(self.rect, style=u'light')
        if self.text == '' and not self.focus:
            self._update_text(self.stub)
        elif self.focus:
            self._update_text(
                self.text[:self.caret_pos] +
                self.prompt +
                self.text[self.caret_pos:]
            )
        else:
            self._update_text(self.text)

    def _insert(self, txt):

        self.text = (
            self.text[:self.caret_pos] +
            txt +
            self.text[self.caret_pos:]
        )
        self.caret_pos += len(txt)

    def coroutine(self):
        r"""Implements the interaction."""
        self.caret_pos = len(self.text)
        retval = None
        while True:
            d = yield retval
            retval = None
            if d[u'type'] == u'stop':
                break
            if d[u'type'] != u'key':
                continue
            key = d[u'key']
            if not self._key_filter(key):
                continue
            if key == u'space':
                self._insert(u' ')
            elif key == u'backspace':
                self.text = (
                    self.text[:self.caret_pos - 1] +
                    self.text[self.caret_pos:]
                )
                self.caret_pos = max(0, self.caret_pos - 1)
            elif key == u'delete':
                self.text = (
                    self.text[:self.caret_pos] +
                    self.text[self.caret_pos + 1:]
                )
            elif key in (u'return', u'enter'):
                if self.return_accepts:
                    retval = self.text
                else:
                    self._insert(u'\n')
            elif key in (u'home', u'page up'):
                self.caret_pos = 0
            elif key in (u'end', u'page down'):
                self.caret_pos = len(self.text)
            elif key == u'left':
                self.caret_pos = max(0, self.caret_pos - 1)
            elif key == u'right':
                self.caret_pos = min(len(self.text), self.caret_pos + 1)
            elif len(key) == 1:
                self._insert(key)
            self._update()
            self.set_var(self.text)


text_input = TextInput



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *


class BaseJoystick:
    """If you insert the JOYSTICK plugin at the start of your experiment, a
    JOYSTICK object automatically becomes part of the experiment object
    and can be used within an INLINE_SCRIPT item as `joystick`.

    {% set arg_joybuttonlist = "A list of buttons that are accepted or " +
    "`None` to accept all buttons." %}
    {% set arg_timeout = "A timeout value in milliseconds or `None` for no " +
    "timeout." %}

    [TOC]
    """
    def __init__(self, experiment, device=0, joybuttonlist=None, timeout=None):
        """Initializes the joystick object.

        Parameters
        ----------
        experiment : experiment
            An Opensesame experiment.
        device : int, optional
            The joystick device number.
        joybuttonlist : list, NoneType, optional
            {{arg_joybuttonlist}}
        timeout : int, float, NoneType, optional
            {{arg_timeout}}
        """
        raise NotImplementedError()

    def set_joybuttonlist(self, joybuttonlist=None):
        r"""Sets a list of accepted buttons.

        Parameters
        ----------
        joybuttonlist : list, NoneType, optional
            {{arg_joybuttonlist}}
        """
        if joybuttonlist is None or joybuttonlist == []:
            self._joybuttonlist = None
        else:
            self._joybuttonlist = []
            for joybutton in joybuttonlist:
                self._joybuttonlist.append(joybutton)

    def set_timeout(self, timeout=None):
        r"""Sets a timeout.

        Parameters
        ----------
        timeout : int, float, NoneType, optional
            {{arg_timeout}}
        """
        self.timeout = timeout

    def get_joybutton(self, joybuttonlist=None, timeout=None):
        r"""Collects joystick button input.

        Parameters
        ----------
        joybuttonlist : list, NoneType, optional
            A list of buttons that are accepted or `None` to default
            joybuttonlist.
        timeout : int, float, NoneType, optional
            A timeout value in milliseconds or `None` to use default timeout.

        Returns
        -------
        tuple
            A (joybutton, timestamp) tuple. The joybutton is `None` if a
            timeout occurs.
        """
        raise NotImplementedError()

    def get_joyaxes(self, timeout=None):
        r"""Waits for joystick axes movement.

        Parameters
        ----------
        timeout : int, float, NoneType, optional
            A timeout value in milliseconds or `None` to use default timeout.

        Returns
        -------
        tuple
            A `(position, timestamp)` tuple. `position` is `None` if a timeout
            occurs. Otherwise, `position` is an `(x, y, z)` tuple.
        """
        raise NotImplementedError()

    def get_joyballs(self, timeout=None):
        r"""Waits for joystick trackball movement.

        Parameters
        ----------
        timeout : int, float, NoneType, optional
            A timeout value in milliseconds or `None` to use default timeout.

        Returns
        -------
        tuple
            A `(position, timestamp)` tuple. The position is `None` if a
            timeout occurs.
        """
        raise NotImplementedError()

    def get_joyhats(self, timeout=None):
        r"""Waits for joystick hat movement.

        Parameters
        ----------
        timeout : int, float, NoneType, optional
            A timeout value in milliseconds or `None` to use default timeout.

        Returns
        -------
        tuple
            A `(position, timestamp)` tuple. `position` is `None` if a timeout
            occurs. Otherwise, `position` is an `(x, y)` tuple.
        """
        raise NotImplementedError()

    def get_joyinput(self, joybuttonlist=None, timeout=None):
        r"""Waits for any joystick input (buttons, axes, hats or balls).

        Parameters
        ----------
        joybuttonlist : list, NoneType, optional
            A list of buttons that are accepted or `None` to default
            joybuttonlist.
        timeout : int, float, NoneType, optional
            A timeout value in milliseconds or `None` to use default timeout.

        Returns
        -------
        tuple
            A (event, value, timestamp) tuple. The value is `None` if a timeout
            occurs. `event` is one of `None`, 'joybuttonpress',
            'joyballmotion', 'joyaxismotion', or 'joyhatmotion'
        """
        raise NotImplementedError()

    def input_options(self):
        r"""Generates a list with the number of available buttons, axes, balls
        and hats.

        Returns
        -------
        list
            A list with number of inputs as: [buttons, axes, balls,
            hats].
        """
        raise NotImplementedError()

    def flush(self):
        r"""Clears all pending input, not limited to the joystick.

        Returns
        -------
        bool
            True if joyinput was pending (i.e., if there was something to
            flush) and False otherwise.
        """
        raise NotImplementedError()



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp.backend import Backend, configurable
import warnings


class Keyboard(Backend):
    """The `Keyboard` class is used to collect keyboard responses. You generally create
    a `Keyboard` object with the `Keyboard()` factory function, as described in the
    section [Creating a Keyboard](#creating-a-keyboard).
    
    __Example__
    
    ~~~ .python
    # Wait for a 'z' or 'x' key with a timeout of 3000 ms
    my_keyboard = Keyboard(keylist=['z', 'x'], timeout=3000)
    start_time = clock.time()
    key, end_time = my_keyboard.get_key()
    response = key
    response_time = end_time - start_time
    ~~~
    
    [TOC]
    
    ## Things to know
    
    ### Creating a Keyboard
    
    You generally create a `Keyboard` with the `Keyboard()` factory function:
    
    ~~~ .python
    my_keyboard = Keyboard()
    ~~~
    
    Optionally, you can pass [Response keywords](#response-keywords) to `Keyboard()`
    to set the default behavior:
    
    ~~~ .python
    my_keyboard = Keyboard(timeout=2000)
    ~~~
    
    ### Key names
    
    - Key names may differ between backends.
    - Keys can be identified either by character or name, and are case-insensitive.
      For example:
      - The key 'a' is represented by 'a' and 'A'
      - The up arrow is represented by 'up' and 'UP'
      - The '/' key is represented by '/', 'slash', and 'SLASH'
      - The spacebar is represented by 'space' and 'SPACE'
    - To find out the name of key, you can:
      - Click on the 'list available keys' button of the KEYBOARD_RESPONSE item.
      - Collect a key press with a KEYBOARD_RESPONSE item, and display the key name
        through a FEEDBACK item with the text 'You pressed [response]' in it.
    
    ### Response keywords
    
    Functions that accept `**resp_args` take the following keyword arguments:
    
    - `timeout` specifies a timeout value in milliseconds, or is set to `None` to
      disable the timeout.
    - `keylist` specifies a list of keys that are accepted, or is set to `None`
      accept all keys.
    
    ~~~ .python
    # Get a left or right arrow press with a timeout of 3000 ms
    my_keyboard = Keyboard()
    key, time = my_keyboard.get_key(keylist=[u'left', u'right'], timeout=3000)
    ~~~
    
    Response keywords only affect the current operation (except when passed to
    `Keyboard()`). To change the behavior for all subsequent
    operations, set the response properties directly:
    
    ~~~ .python
    # Get two key A or B presses with a 5000 ms timeout
    my_keyboard = Keyboard()
    my_keyboard.keylist = [u'a', u'b']
    my_keyboard.timeout = 5000
    key1, time1 = my_keyboard.get_key()
    key2, time2 = my_keyboard.get_key()
    ~~~
    
    Or pass the response options to [keyboard.__init__][__init__]:
    
    ~~~ .python
    # Get two key A or B presses with a 5000 ms timeout
    my_keyboard = Keyboard(keylist=[u'a', u'b'], timeout=5000)
    key1, time1 = my_keyboard.get_key()
    key2, time2 = my_keyboard.get_key()
    ~~~
    """
    def __init__(self, experiment, **resp_args):
        r"""Constructor to create a new `Keyboard` object. You do not generally
        call this constructor directly, but use the `Keyboard()` function,
        which is described here: [/python/common/]().

        Parameters
        ----------
        experiment : experiment
            The experiment object.
        **resp_args : dict
            Optional [response keywords](#response-keywords) (`timeout` and
            `keylist`) that will be used as the default for this `Keyboard`
            object.


        Examples
        --------
        >>> my_keyboard = Keyboard(keylist=['z', 'm'], timeout=2000)
        """
        self.experiment = experiment
        Backend.__init__(self, configurables={
            u'timeout': self.assert_numeric_or_None,
            u'keylist': self.assert_list_or_None,
        }, **resp_args)

    def set_config(self, **cfg):

        # Create a new keylist that also contains synonyms
        if u'keylist' in cfg and isinstance(cfg[u'keylist'], list):
            keylist = list(cfg[u'keylist'])
            for key in keylist:
                keylist += [key for key in self.synonyms(key)
                            if key not in keylist]
            cfg[u'keylist'] = keylist
        Backend.set_config(self, **cfg)

    def default_config(self):

        return {
            u'timeout': None,
            u'keylist': None,
        }

    @configurable
    def get_key(self, **resp_args):
        r"""Collects a single key press.

        Parameters
        ----------
        **resp_args : dict
            Optional [response keywords](#response-keywords) (`timeout` and
            `keylist`) that will be used for this call to `Keyboard.get_key()`.
            This does not affect subsequent operations.


        Returns
        -------
        tuple
            A `(key, timestamp)` tuple. `key` is None if a timeout occurs.

        Examples
        --------
        >>> my_keyboard = Keyboard()
        >>> response, timestamp = my_keyboard.get_key(timeout=5000)
        >>> if response is None:
        >>>         print(u'A timeout occurred!')
        """
        raise NotImplementedError()

    @configurable
    def get_key_release(self, **resp_args):
        r"""*New in v3.2.0*

        Collects a single key release.

        *Important:* This
        function currently assumes a QWERTY keyboard
        layout (unlike
        `Keyboard.get_key()`). This means that the returned
        `key` may be
        incorrect on non-QWERTY keyboard layouts. In addition,
        this function is
        not implemented for the *psycho* backend.

        Parameters
        ----------
        **resp_args : dict
            Optional [response keywords](#response-keywords) (`timeout` and
            `keylist`) that will be used for this call to
            `Keyboard.get_key_release()`. This does not affect subsequent
            operations.

        Returns
        -------
        tuple
            A `(key, timestamp)` tuple. `key` is None if a timeout occurs.

        Examples
        --------
        >>> my_keyboard = Keyboard()
        >>> response, timestamp = my_keyboard.get_key_release(timeout=5000)
        >>> if response is None:
        >>>         print(u'A timeout occurred!')
        """
        raise NotImplementedError()

    def get_mods(self):
        r"""Returns a list of keyboard moderators (e.g., shift, alt, etc.) that
        are currently pressed.

        Returns
        -------
        list
            A list of keyboard moderators. An empty list is returned if no
            moderators are pressed.

        Examples
        --------
        >>> my_keyboard = Keyboard()
        >>> moderators = my_keyboard.get_mods()
        >>> if u'shift' in moderators:
        >>>         print(u'The shift-key is down!')
        """
        raise NotImplementedError()

    def valid_keys(self):
        r"""Tries to guess which key names are accepted by the back-end. For
        internal use.

        Returns
        -------
        list
            A list of valid key names.
        """
        raise NotImplementedError()

    def synonyms(self, key):
        r"""Gives a list of synonyms for a key, either codes or names. Synonyms
        include all variables as types and as Unicode strings (if applicable).

        Returns
        -------
        list
            A list of synonyms
        """
        raise NotImplementedError()

    def flush(self):
        r"""Clears all pending keyboard input, not limited to the keyboard.

        Returns
        -------
        bool
            True if a key had been pressed (i.e., if there was something to
            flush) and False otherwise.
        """
        raise NotImplementedError()

    def show_virtual_keyboard(self, visible=True):
        r"""Shows or hides a virtual keyboard if this is supported by the
        back-end. This function is only necessary if you want the virtual
        keyboard to remain visible while collecting multicharacter
        responses. Otherwise, `Keyboard.get_key()` will implicitly show and
        hide the keyboard for a single-character response.

        This function does nothing for back-ends that do not support virtual
        keyboards.

        Parameters
        ----------
        visible : bool, optional
            True if the keyboard should be shown, False otherwise.

        Examples
        --------
        >>> my_keyboard = Keyboard()
        >>> my_keyboard.show_virtual_keyboard(True)
        >>> response1, timestamp2 = my_keyboard.get_key()
        >>> response2, timestamp2 = my_keyboard.get_key()
        >>> my_keyboard.show_virtual_keyboard(False)
        """
        pass

    def _keycode_to_str(self, keycode):
        """visible:	False"""
        raise NotImplementedError()


# Non PEP-8 alias for backwards compatibility
keyboard = Keyboard



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *


class ItemStack:
    """Keeps track of which item is currently active."""
    
    def __init__(self):
        self.clear()

    def clear(self):
        """Clears the stack."""
        self._stack = []

    def push(self, item, phase):
        """Adds a new item to the stack.

        Parameters
        ----------
        item : str
            The item to add.
        """
        self._stack.append((item, phase))

    def pop(self):
        """Pops the last item from the stack.

        Returns
        -------
        str
            The popped item.
        """
        return self._stack.pop()

    def __str__(self):

        return '.'.join(['%s[%s]' % i for i in self._stack])
        
    def __getitem__(self, key):
        return self._stack[key]


# Create a single instance of the stack
item_stack_singleton = ItemStack()



# -*- coding:utf-8 -*-

"""
This file is part of openexp.

openexp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

openexp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with openexp.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
from libopensesame.py3compat import *
from libqtopensesame.sketchpad_elements._base_element import BaseElement
from libopensesame.sketchpad_elements import Textline as TextlineRuntime
from libqtopensesame.dialogs.text_input import TextInput
from qtpy import QtCore, QtWidgets
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'sketchpad', category=u'item')


class Textline(BaseElement, TextlineRuntime):

    r"""A textline element.
    See base_element for docstrings and function
    descriptions.
    """
    def show_edit_dialog(self):
        r"""The show-edit dialog for the textline only edits the text, not the
        full element script.
        """
        content = safe_decode(self.properties['text']).replace('<br />', '\n')
        text = TextInput(self.main_window,
                         msg=_('Please enter a text for the textline'),
                         content=content).get_input()
        if text is None:
            return
        self.properties[u'text'] = self.clean_text(text)
        self.sketchpad.draw()

    @classmethod
    def mouse_press(cls, sketchpad, pos):
        text = TextInput(sketchpad.main_window,
                         msg=_('Enter text')).get_input()
        if text is None:
            return None
        properties = {
            u'x':			pos[0],
            u'y':			pos[1],
            u'text':		cls.clean_text(text, escape=True),
            u'color': 		sketchpad.current_color(),
            u'center': 		sketchpad.current_center(),
            u'font_family': sketchpad.current_font_family(),
            u'font_size': 	sketchpad.current_font_size(),
            u'font_bold': 	sketchpad.current_font_bold(),
            u'font_italic': sketchpad.current_font_italic(),
            u'html':		sketchpad.current_html(),
            u'show_if': 	sketchpad.current_show_if()
        }
        return Textline(sketchpad, properties=properties)

    @staticmethod
    def clean_text(text, escape=False):
        r"""Cleans text by removing quotes and converting newlines to <br />
        tags.

        Parameters
        ----------
        text
            The text to clean.
        typeescape, optional
            Indicates whether slashes should be escaped.
        type, optional
            bool

        Returns
        -------
        unicode
            Clean text.
        """
        text = str(text)
        text = text.replace(os.linesep, u'<br />')
        text = text.replace(u'\n', u'<br />')
        text = text.replace(u'"', u'')
        if escape:
            text = text.replace(u'\\', u'\\\\')
        return text

    @staticmethod
    def requires_text():
        return True

    @staticmethod
    def requires_color():
        return True

    @staticmethod
    def requires_center():
        return True


# Alias for backwards compatibility
textline = Textline



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from qtpy.QtWidgets import QTabWidget
from libqtopensesame.misc.config import cfg
from .jupyter_console import JupyterConsole
from .console_cornerwidget import ConsoleCornerWidget
from .constants import KERNEL_NAMES


class ConsoleTabWidget(QTabWidget, BaseSubcomponent):

    def __init__(self, parent, kwargs={}):

        super(ConsoleTabWidget, self).__init__(parent)
        self.setup(parent)
        self._console_count = 1
        self._kwargs = kwargs
        self.setCornerWidget(ConsoleCornerWidget(self, kwargs))
        if cfg.jupyter_inprocess:
            # Also start a single inprocess console, mostly for debugging
            # purposes
            self.add(inprocess=True)
        self.tabCloseRequested.connect(self.close)
        self.currentChanged.connect(self._on_switch)
        self.add()

    def add(self, inprocess=False, **kwargs):

        merged_kwargs = self._kwargs.copy()
        merged_kwargs.update(kwargs)
        kernel = merged_kwargs.get('kernel', cfg.jupyter_default_kernel)
        name = u'{} ({})'.format(
            KERNEL_NAMES.get(kernel, kernel),
            self._console_count
        )
        jupyter_console = JupyterConsole(
            self,
            name=name,
            inprocess=inprocess,
            **merged_kwargs
        )
        self.addTab(
            jupyter_console,
            self.main_window.theme.qicon(
                'utilities-terminal'
                if inprocess
                else 'os-debug'
            ),
            name
        )
        if inprocess:
            global_dict = {
                ext.name(): ext
                for ext in self.extension_manager.extensions
            }
            global_dict['opensesame'] = self.main_window
            global_dict['event_durations'] = self._event_durations
            jupyter_console.set_workspace_globals(global_dict)
        else:
            self.extension_manager.fire(
                'register_subprocess',
                pid=jupyter_console.pid,
                description='jupyter_console:{}'.format(name)
            )
        self.setTabsClosable(self.count() > 1)
        self.setCurrentIndex(self.count() - 1)
        self._console_count += 1
        self.extension_manager.fire(
            'workspace_new',
            name=name,
            workspace_func=jupyter_console.get_workspace_globals
        )

    def close(self, index):

        console = self.widget(index)
        console.shutdown()
        self.removeTab(index)
        self.setTabsClosable(self.count() > 1)
        self.extension_manager.fire(
            'workspace_close',
            name=console.name
        )

    def close_all(self):

        while self.count():
            self.close(0)

    @property
    def current(self):

        return self.currentWidget()

    def _on_switch(self, index):

        if self.widget(index) is None:
            return
        self.extension_manager.fire(
            'workspace_switch',
            name=self.widget(index).name,
            workspace_func=self.widget(index).get_workspace_globals
        )

    def _event_durations(self):
        
        """Registered as a global function in the Python workspace to get an
        overview of event durations for development purposes.
        """
        
        if not self.main_window.options.profile:
            print('Start with --performance-profile to see event durations')
            return
        event_durations = []
        for ext in self.extension_manager._extensions:
            for event, durations in ext._event_durations.items():
                event_durations.append(
                    (ext, event, sum(durations) / len(durations))
                )
        for ext, event, duration in sorted(
            event_durations, key=lambda e: -e[2]
        ):
            print('{}({}): {:.2f} ms'.format(event, ext.name(), duration))



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from libqtopensesame.widgets.tree_base_item import TreeBaseItem
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'tree_general_item', category=u'core')


class TreeGeneralItem(TreeBaseItem):

    r"""Corresponds to the general widget in the overview area."""
    def __init__(self, main_window):
        r"""Constructor.

        Parameters
        ----------
        main_window : qtopensesame
            The main-window object.
        """
        super().__init__()
        self.setup(main_window)
        self.setText(0, safe_decode(self.experiment.var.title))
        self.setIcon(0, self.theme.qicon(u'os-experiment'))
        self.setToolTip(0, _(u'General options'))
        self._droppable = False
        self._draggable = False
        self.name = u'__general__'
        if self.experiment.var.start in self.experiment.items:
            try:
                self.experiment.items[self.experiment.var.start] \
                    .build_item_tree(self)
            except RecursionError:
                from libqtopensesame.widgets.tree_recursion_error_item import (
                    tree_recursion_error_item
                )
                self.addChild(
                    tree_recursion_error_item(self)
                )
            self.child(0).set_draggable(False)
        self.expand()

    def ancestry(self):

        return u'__general__', u'__general__'

    def open_tab(self):

        self.main_window.tabwidget.open_general()


# Alias for backwards compatibility
tree_general_item = TreeGeneralItem



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
from libqtopensesame.extensions import BaseExtension
from libopensesame.experiment import Experiment


class ExampleExperiments(BaseExtension):

    def event_startup(self):
        self.initialized = False

    def add_experiment(self, path):
        r"""Processes one experiment.

        Parameters
        ----------
        path : str
            The experiment folder name.
        """
        src = os.path.join(self.example_folder, path, u'experiment.osexp')
        if not os.path.exists(src):
            return
        try:
            e = Experiment(u'dummy', src, experiment_path=os.path.dirname(src))
        except Exception as e:
            oslogger.warning(f'failed to load experiment: {e}')
            return
        d = e.front_matter
        # We need to use forward slashes, also on Windows, otherwise it leads
        # to an invalid URL when prefixed with opensesame://
        d[u'src'] = src.replace(u'\\', u'/')
        if u'Title' not in d:
            d[u'Title'] = u'No title'
        if u'Description' not in d:
            d[u'Description'] = u'No description'
        self.experiments.append(d)

    def initialize(self):
        r"""Initializes the plug-in by building a Markdown page with examples.
        """
        self.set_busy()
        self.example_folder = os.path.join(self.extension_folder, 'examples')
        self.experiments = []
        for path in sorted(os.listdir(self.example_folder)):
            self.add_experiment(path)
        self.generate_markdown()
        self.set_busy(False)

    def generate_markdown(self):
        r"""Generates a Markdown string of all experiment info."""
        self.md = u'# Example experiments\n\n'
        for d in self.experiments:
            self.md += (
                u'[%(Title)s](opensesame://%(src)s)  \n'
                u'*%(Description)s*  \n'
            ) % d
            if u'Citation' in d:
                self.md += u'\n* <small>%s</small>\n' % d[u'Citation']
            self.md += u'\n<hr />\n'

    def activate(self):
        r"""Shows the list of example experiments."""
        if not self.initialized:
            self.initialize()
        self.tabwidget.open_markdown(self.md, icon=u'help-contents',
                                     title=u'Example experiments')



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import OSException, ItemDoesNotExist, InvalidValue
from libopensesame.base_response_item import BaseResponseItem
from openexp._canvas.canvas import Canvas
from openexp.mouse import Mouse


class MouseResponseMixin:

    r"""A mixin class that should be inherited along with base_response_item by
    all classes that want to collect mouse responses.
    """
    _resp_codes = {
        u'none': None,
        u'timeout': None,
        u'left_button': 1,
        u'middle_button': 2,
        u'right_button': 3,
        u'left': 1,
        u'middle': 2,
        u'right': 3,
        u'scroll_up': 4,
        u'scroll_down': 5
    }

    def button_code(self, response):

        if response is None or isinstance(response, int):
            return response
        try:
            return int(response)
        except:
            return self._resp_codes[response.lower()]

    def prepare_response_func(self):
        """See base_response_item."""
        if self._allowed_responses is None:
            buttonlist = None
        else:
            buttonlist = [self.button_code(r) for r in self._allowed_responses]
        self._mouse = Mouse(
            self.experiment,
            timeout=self._timeout,
            buttonlist=buttonlist
        )
        if self.var.get(u'event_type', default=u'mouseclick') == u'mouseclick':
            return self._mouse.get_click
        if self.var.event_type == u'mouserelease':
            return self._mouse.get_click_release
        raise InvalidValue(
            f'event_type should be "mouseclick" or "mouserelease", not '
            f'{self.var.event_type}')

    def process_response(self, response_args):
        """See base_response_item."""
        response, pos, t1 = response_args
        if pos is None:
            self.experiment.var.cursor_x = u'NA'
            self.experiment.var.cursor_y = u'NA'
        else:
            self.experiment.var.cursor_x, self.experiment.var.cursor_y = pos
        if self.var.get(u'linked_sketchpad', default=u'') and pos is not None:
            if self.var.linked_sketchpad not in self.experiment.items:
                raise ItemDoesNotExist(self.var.linked_sketchpad)
            item = self.experiment.items[self.var.linked_sketchpad]
            if not hasattr(item, u'canvas') or \
                    not isinstance(item.canvas, Canvas):
                raise OSException(
                    f'Linked sketchpad does not have a canvas: '
                    f'{self.var.linked_sketchpad}')
            self.experiment.var.cursor_roi = u';'.join(
                item.canvas.elements_at(*pos)
            )
        else:
            self.experiment.var.cursor_roi = u'undefined'
        BaseResponseItem.process_response(self, (response, t1))


class MouseResponse(MouseResponseMixin, BaseResponseItem):

    r"""An item for collecting mouse responses."""
    description = u'Collects mouse responses'
    process_feedback = True

    def reset(self):
        """See item."""
        self.var.flush = u'yes'
        self.var.show_cursor = u'yes'
        self.var.timeout = u'infinite'
        self.var.duration = u'mouseclick'
        self.var.linked_sketchpad = u''
        self.var.event_type = u'mouseclick'
        self.var.unset(u'allowed_responses')
        self.var.unset(u'correct_response')

    def validate_response(self, response):
        """See base_response_item."""
        try:
            self.button_code(response)
        except:
            return False
        return True

    def response_matches(self, test, ref):
        """See base_response_item."""
        return any(self.button_code(test) == self.button_code(r) for r in ref)

    def prepare(self):
        """See item."""
        super().prepare()
        self._flush = self.var.flush == u'yes'

    def run(self):
        """See item."""
        if self._flush:
            self._mouse.flush()
        # Show cursor if necessary
        if self.var.show_cursor == u'yes':
            self._mouse.visible = True
        super().run()
        self._mouse.visible = False

    def coroutine(self):
        """See coroutines plug-in."""
        self._mouse.timeout = 0
        alive = True
        yield
        self._t0 = self.set_item_onset()
        if self._flush:
            self._mouse.flush()
        if self.var.show_cursor == u'yes':
            self._mouse.visible = True
        while alive:
            button, pos, time = self._mouse.get_click()
            if button is not None:
                break
            alive = yield
        self.process_response((button, pos, time))
        self._mouse.visible = False

    def var_info(self):
        """See item."""
        return super().var_info() + [
            (u'cursor_x', u'[Depends on response]'),
            (u'cursor_y', u'[Depends on response]'),
            (u'cursor_roi', u'[Depends on response]')
        ]


# Alias for backwards compatibility
mouse_response = MouseResponse
mouse_response_mixin = MouseResponseMixin



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import numpy as np
from openexp._canvas import canvas


class PsychoElement:

    r"""Together with Element, PsychoElement is the base object for all psycho
    sketchpad elements.
    """
    @property
    def win(self):
        return self._canvas.experiment.window

    def show(self):

        if self.visible:
            self._stim.draw()

    def _on_attribute_change(self, **kwargs):

        if self._canvas.auto_prepare:
            self.prepare()

    def _mask(self, env, size, stdev):
        r"""Generates a PsychoPy mask for Gabor and NoisePatch stimuli.

        Parameters
        ----------
        env : str
            The envelope.
        size : int
            The stimulus size.
        stdev : int
            The standard deviation of the mask if the envelope is gaussian.

        Returns
        -------
        ndarray
            A PsychoPy mask, which is a numpy array.
        """
        # Get the smallest power-of-two that is larger than or equal to the
        # given size
        size = int(np.ceil(np.sqrt(size))**2)
        # Create a PsychoPy mask
        env = canvas._match_env(env)
        if env == u'c':
            return u'circle', size
        if env == u'g':
            return u'gauss', 6 * stdev
        if env == u'r':
            return u'None', size
        if env == u'l':
            _env = np.zeros([size, size])
            for x in range(size):
                for y in range(size):
                    r = np.sqrt((x-size/2)**2+(y-size/2)**2)
                    _env[x, y] = (max(0, (0.5*size-r) / (0.5*size))-0.5)*2
            return _env, size
        raise ValueError('Invalid mask')


class RotatingElement:

    def _on_attribute_change(self, **kwargs):

        if u'rotation' in kwargs:
            self._stim.ori = kwargs.pop(u'rotation')
        if kwargs:
            super(RotatingElement, self)._on_attribute_change(**kwargs)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame. If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.base_response_item import BaseResponseItem
from openexp.keyboard import Keyboard


class Joystick(BaseResponseItem):

    process_feedback = True

    def reset(self):
        self.var.timeout = u'infinite'
        self.var.allowed_responses = u''
        self.var._dummy = u'no'
        self.var._device = 0

    def validate_response(self, response):
        try:
            response = int(response)
        except ValueError:
            return False
        return response >= 0 or response <= 255

    def _get_button_press(self):
        r"""Calls libjoystick.get_button_press() with the correct arguments."""
        return self.experiment.joystick.get_joybutton(
            joybuttonlist=self._allowed_responses,
            timeout=self._timeout
        )

    def prepare_response_func(self):
        self._keyboard = Keyboard(
            self.experiment,
            keylist=(
                self._allowed_responses if self._allowed_responses
                else list(range(0, 10))  # Only numeric keys
            ),
            timeout=self._timeout
        )
        if self.var._dummy == u'yes':
            return self._keyboard.get_key
        # Dynamically load a joystick instance
        if not hasattr(self.experiment, u'joystick'):
            from .libjoystick import LibJoystick
            self.experiment.joystick = LibJoystick(self.experiment, 
                                                   device=self.var._device)
            self.python_workspace[u'joystick'] = self.experiment.joystick
        if self._allowed_responses is not None:
            self._allowed_responses = [int(r) for r in self._allowed_responses]
        return self._get_button_press

    def response_matches(self, test, ref):
        return safe_decode(test) in ref

    def coroutine(self):
        if self.var._dummy == u'yes':
            self._keyboard.timeout = 0
        else:
            self._timeout = 0
        alive = True
        yield
        self._t0 = self.set_item_onset()
        while alive:
            button, time = self._collect_response()
            if button is not None:
                break
            alive = yield
        self.process_response((button, time))



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import InvalidValue
from libopensesame.widgets._widget import Widget
from openexp.canvas_elements import RichText


class RatingScale(Widget):

    r"""The rating_scale widget is a horizontally aligned series of checkable
    boxes (nodes), optionally with a label attached to each node.

    __Example
    (OpenSesame script):__

    ~~~
    widget 0 0 1 1 label text="I like fluffy
    kittens"
    widget 0 1 1 1 rating_scale var="response" nodes="Agree;Don't
    know;Disagree"
    ~~~

    __Example (OpenSesame script):__

    ~~~ .python
    form =
    Form()
    label = Label(text='I like fluffy kittens')
    rating_scale =
    RatingScale(nodes=['Agree', "Don't know",
            'Disagree'],
    var='response')
    form.set_widget(label, (0,0))
    form.set_widget(rating_scale,
    (0,1))
    form._exec()
    ~~~

    [TOC]
    """
    def __init__(self, form, nodes=5, click_accepts=False,
                 orientation=u'horizontal', var=None, default=None):
        r"""Constructor to create a new `RatingScale` object. You do not
        generally call this constructor directly, but use the
        `RatingScale()`
        factory function, which is described here:
        [/python/common/]().

        Parameters
        ----------
        form : form
            The parent form.
        nodes : int, list, optional
            The number of nodes or a list of node identifiers (e.g., ['yes',
            'no', 'maybe']. If a list is passed the rating scale will have
            labels, otherwise it will just have boxes.
        click_accepts : bool, optional
            Indicates whether the form should close when a value is selected.
        orientation : str, unicode, optional
            'horizontal' indicates a horizontally oriented rating
            scale,
            'vertical' indicates a vertically oriented rating
            scale.
        var : str, unicode, NoneType, optional
            The name of the experimental variable that should be used to log
            the widget status. The value that is logged is the number of the
            node that was selected, with the first node being 0. If no nodes
            were selected, the value is 'None'. For more information about the
            use of response variables in forms, see the form documentation
            page.
        default : int, NoneType, optional
            The node that is selected by default, or `None` to select no node.
            The value corresponds to the node number, where 0 is the first
            node.
        """
        if isinstance(click_accepts, str):
            click_accepts = click_accepts == u'yes'
        Widget.__init__(self, form)
        self.type = u'rating_scale'
        self.click_accepts = click_accepts
        self.pos_list = []
        self.var = var
        self.orientation = orientation
        if type(nodes) == int:
            self.nodes = [u'']*nodes
        elif isinstance(nodes, str):
            self.nodes = nodes.split(u';')
        else:
            self.nodes = nodes
        self.set_value(default)

    def on_mouse_click(self, pos):
        r"""Is called whenever the user clicks on the widget. Selects the
        correct value from the scale and optionally closes the form.

        Parameters
        ----------
        pos : tuple
            An (x, y) coordinates tuple.
        """
        for i, (x, y) in enumerate(self.pos_list):
            if not self._inside(pos, (x, y, self.box_size, self.box_size)):
                continue
            self.set_value(i)
            self._update()
            if self.click_accepts:
                return i
            break

    def _init_canvas_elements(self):
        r"""Initializes all canvas elements."""
        x, y, w, h = self.rect
        cx = x+w/2
        cy = y+h/2
        bs = self.form.theme_engine.box_size()
        # The outline
        if self.orientation == 'horizontal':
            box_rect = x, cy-.5*bs, w, 2*bs
        elif self.orientation == 'vertical':
            box_rect = cx-.5*bs, y, 2*bs, h
        else:
            raise InvalidValue(f'rating_scale orientation must be '
                               f'"horizontal" or "vertical", not '
                               f'{self.orientation}')
        self.canvas.add_element(
            self.form.theme_engine.frame(*box_rect, style=u'light')
        )
        self._checked_boxes = []
        self._unchecked_boxes = []
        # The distances between nodes
        dx = (1*w-3*bs)/(len(self.nodes)-1)
        dy = (1*h-3*bs)/(len(self.nodes)-1)
        for i, node in enumerate(self.nodes):
            text_width, text_height = RichText(
                node).construct(self.canvas).size
            if self.orientation == 'horizontal':
                node_x = x + bs + i * dx
                node_y = cy
                text_x = node_x + self.box_size / 2
                text_y = cy - bs - .5 * text_height
                center = True
            elif self.orientation == 'vertical':
                node_x = cx
                node_y = y + bs + i * dy
                text_x = cx + 2 * bs
                text_y = node_y - .5 * text_height + .5 * bs
                center = False
            self.pos_list.append((node_x, node_y))
            self.canvas.add_element(
                RichText(node, center=center, x=text_x, y=text_y)
            )
            cb = self.form.theme_engine.box(node_x, node_y, checked=True)
            ub = self.form.theme_engine.box(node_x, node_y, checked=False)
            self._checked_boxes.append(cb)
            self._unchecked_boxes.append(ub)
            self.canvas.add_element(cb)
            self.canvas.add_element(ub)

    def _update(self):
        r"""Draws the widget."""
        for i, (cb, ub) in enumerate(
                zip(self._checked_boxes, self._unchecked_boxes)):
            cb.visible = i == self.value
            ub.visible = i != self.value

    def set_value(self, val):
        r"""Sets the rating scale value.

        Parameters
        ----------
        val : int
            The value.
        """
        if val is not None and (val >= len(self.nodes) or val < 0):
            raise InvalidValue(f'Trying to select a non-existing node ({val}).'
                               f'Did you specify an incorrect default value?')
        self.value = val
        self.set_var(val)


rating_scale = RatingScale



# -*- coding:utf-8 -*-

"""
This file is part of openexp.

openexp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

openexp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with openexp.  If not, see <http://www.gnu.org/licenses/>.
"""
import math
from libopensesame.py3compat import *
from libqtopensesame.misc.config import cfg
from libqtopensesame.sketchpad_elements._base_element import BaseElement
from libopensesame.sketchpad_elements import Circle as CircleRuntime


class Circle(BaseElement, CircleRuntime):

    pos_start = None
    preview = None

    @classmethod
    def mouse_press(cls, sketchpad, pos):

        if cls.pos_start is not None:
            return
        cls.pos_start = pos
        xc = sketchpad.canvas.xcenter()
        yc = sketchpad.canvas.ycenter()
        cls.preview = sketchpad.canvas.circle(pos[0], pos[1], 0,
            color=cfg.sketchpad_preview_color,
            penwidth=cfg.sketchpad_preview_penwidth)

    @classmethod
    def mouse_release(cls, sketchpad, pos):

        if cls.pos_start is None:
            cls.pos_start = pos
            return None
        if cls.radius(pos) < 1:
            cls.pos_start = None
            sketchpad.canvas.removeItem(cls.preview)
            return
        properties = {
            u'x':		cls.pos_start[0],
            u'y':		cls.pos_start[1],
            u'r':		cls.radius(pos),
            u'color': 	sketchpad.current_color(),
            u'penwidth': sketchpad.current_penwidth(),
            u'fill': sketchpad.current_fill(),
            u'show_if': sketchpad.current_show_if()
        }
        e = Circle(sketchpad, properties=properties)
        cls.pos_start = None
        sketchpad.canvas.removeItem(cls.preview)
        return e

    @classmethod
    def mouse_move(cls, sketchpad, pos):

        if cls.pos_start is None or cls.preview is None:
            return
        xc = sketchpad.canvas.xcenter()
        yc = sketchpad.canvas.ycenter()
        r = cls.radius(pos)
        x = cls.pos_start[0]-r
        y = cls.pos_start[1]-r
        cls.preview.setRect(x, y, 2*r, 2*r)

    @classmethod
    def radius(cls, pos):

        return math.sqrt((pos[0]-cls.pos_start[0])**2 +
                         (pos[1]-cls.pos_start[1])**2)

    @classmethod
    def reset(cls):

        cls.pos_start = None

    @staticmethod
    def click(sketchpad, pos):

        global pos_start
        if pos_start is None:
            pos_start = pos
            return None
        r = cls.radius(pos)
        properties = {
            u'x':		pos_start[0],
            u'y':		pos_start[1],
            u'r':		r,
            u'color': 	sketchpad.current_color(),
            u'penwidth': sketchpad.current_penwidth(),
            u'fill': sketchpad.current_fill(),
            u'show_if': sketchpad.current_show_if()
        }
        e = circle(sketchpad, properties=properties)
        pos_start = None
        return e

    @staticmethod
    def requires_color():
        return True

    @staticmethod
    def requires_penwidth():
        return True

    @staticmethod
    def requires_fill():
        return True


# Alias for backwards compatibility
circle = Circle



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from qtpy.QtGui import QColor, QBrush
from libqtopensesame.widgets.tree_base_item import TreeBaseItem
from libqtopensesame.items.qtstructure_item import QtStructureItem
from libqtopensesame.misc.translate import translation_context
_ = translation_context('tree_item_item', category='core')


class TreeItemItem(TreeBaseItem):
    """Corresponds to an item widget in the overview area.

    Parameters
    ----------
    item : qtitem
        An item.
    extra_info : None or unicode, optional
        Extra info that is shown in the second column. Not shown in
        overview mode.
    """
    def __init__(self, item, extra_info=None):
        super().__init__()
        self.setup(item.main_window)
        self.item = item
        tooltip = _(u"Type: %s\nDescription: %s") % (item.item_type,
                                                     item.var.description)
        self.setText(0, item.name)
        self.set_extra_info(extra_info)
        self.setFlags(QtCore.Qt.ItemIsEditable | self.flags())
        self.setIcon(0, self.theme.qicon(item.item_icon()))
        self.name = item.name
        # Only allow drops on used items
        self._droppable = self.item.name in self.item.experiment.items.used()
        self._draggable = True
        self._lock = False
        self.setToolTip(0, tooltip)

    @property
    def open_tab(self):
        return self.item.open_tab

    @property
    def close_tab(self):
        return self.item.close_tab

    def has_append_menu(self):
        return self.item.item_type == 'sequence'

    def ancestry(self):
        """Gets the full ancestry of a tree item, i.e. a sequence of items that
        are above the item in the hierarchy. The index of the item in the
        parent is indicated by a ':'. The index is 0 in the case of most items,
        but is mostly necessary for indicating the position in sequence items.

        For example:

            fixdot:2.trial_sequence:0.block_loop:0.experiment:0
            
        Returns
        -------
        tuple
            A (item name, ancestry) tuple. For example:
            ('trial_sequence', 'trial_sequence:0.block_loop:0.experiment:0')
        """
        treeitem = self
        item_name = str(treeitem.text(0))
        l = []
        while True:
            if treeitem.parent() is not None:
                index = treeitem.parent().indexOfChild(treeitem)
            else:
                index = 0
            l.append(str(treeitem.text(0))+':'+str(index))
            treeitem = treeitem.parent()
            if treeitem is None or not treeitem.droppable:
                break
        return item_name, '.'.join(l)

    def show_context_menu(self, pos):
        """Pops up the item context menu.

        Parameters
        ----------
        pos : QPoint
            The cursor position.
        """
        from libqtopensesame.widgets.item_context_menu import item_context_menu
        menu = item_context_menu(self.main_window, self)
        menu.popup(pos)

    def rename(self, from_name, to_name):
        """Renames an item.

        Parameters
        ----------
        from_name : unicode
            The old item name.
        to_name : unicode
            The new item name.
        """
        super().rename(from_name, to_name)
        if str(self.text(0)) == from_name:
            self.setText(0, to_name)
            self.name = to_name

    def start_rename(self):
        """Goes into edit mode for the item's name."""
        self.treeWidget().editItem(self, 0)

    def start_edit_runif(self):
        """Goes into edit mode for the item's run-if statement. This is only
        applicable to sequences, i.e. not if the treewidget is in overview
        mode.
        """
        if not self.treeWidget().overview_mode:
            self.treeWidget().editItem(self, 1)

    def set_icon(self, name, icon):
        super().set_icon(name, icon)
        if str(self.text(0)) == name:
            self.setIcon(0, self.theme.qicon(icon))

    def drop_hint(self):
        if self.treeWidget().overview_mode or self.parent() is None:
            if self.item.item_type == 'loop':
                return _('Set as item to run for %s') % self.name
            if self.item.item_type == 'sequence':
                return _('Insert into %s') % self.name
        return _('Drop below %s') % self.name

    def is_deletable(self):
        """
        Returns
        -------
        bool
            True if the item for this treeitem can be deleted, False otherwise.
        """
        return hasattr(self.parent(), 'item')

    def is_unused(self):
        """
        Returns
        -------
        bool
            True if the item is unused, False otherwise.
        """
        return self.parent() is not None and self.parent().name == '__unused__'

    def is_cloneable(self):
        """
        Returns
        -------
        bool
            True if the item for this treeitem can be cloned, False otherwise.
            An item can be cloned if it is in a sequence.
        """
        if not hasattr(self.parent(), 'item'):
            return False
        if not getattr(self.parent(), 'item').item_type == 'sequence':
            return False
        return True
    
    def disable(self):
        self._set_enabled(False)

    def enable(self):
        self._set_enabled(True)
        
    def _set_enabled(self, enabled):
        index = self.parent().indexOfChild(self)
        parent_item = self.parent().item
        parent_item.enable_child_item(self.item.name, index, enabled)
        parent_item.update()
        self.experiment.build_item_tree()

    def delete(self):
        """Deletes the item, if possible."""
        if not self.is_deletable():
            return
        index = self.parent().indexOfChild(self)
        parent_item = self.parent().item
        parent_item.remove_child_item(self.item.name, index)
        parent_item.update()
        self.experiment.build_item_tree()

    def permanently_delete(self):
        """Permanently deletes the item, if possible."""
        if not self.is_deletable() and not self.is_unused():
            return
        if QtWidgets.QMessageBox.question(
                self.treeWidget(), 
                _('Permanently delete item'), 
                _('Are you sure you want to permanently delete <b>%s</b>? All linked copies of <b>%s</b> will be deleted. You will not be able to undo this.')
                  % (self.name, self.name),
                buttons=(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No), 
                defaultButton=QtWidgets.QMessageBox.No
            ) != QtWidgets.QMessageBox.Yes:
            return
        del self.item_store[self.name]
        self.close_tab()
        try:
            # Sometimes the treeWidget has already been deleted, in which case
            # this gives an Exception. But it when it hasn't been deleted, a
            # structure_change has to be emitted, otherwise the GUI doesn't
            # update properly.
            self.treeWidget().structure_change.emit()
        except Exception as e:
            pass

    def copy_unlinked(self):
        """Copies a snippet of the current item plus children to the keyboard.
        """
        import json

        data = {'type': 'item-snippet',
                'main-item-name': self.item.name,
                'items': []}
        for item_name in [self.item.name] \
                + self.experiment.items[self.item.name].children():
            # Don't create multiple linked copies of the same object. This
            # avoids for example, a sequence with two copies of the same item
            # from being copied as a sequence with two disconnected items. In
            # general, the user will want to preserve linked copies at this
            # level.
            if any(item_name == item_dict['item-name']
                   for item_dict in data['items']):
                continue
            item = self.experiment.items[item_name]
            data['items'].append({
                'item-name': item_name,
                'item-type': item.item_type,
                'script': item.to_string()
            })
        text = safe_decode(json.dumps(data))
        QtWidgets.QApplication.clipboard().setText(text)

    def copy_linked(self):
        """Copies a linked copy to the keyboard"""
        import json

        data = {
            'type': 'item-existing',
            'item-name': self.item.name,
            'item-type': self.item.item_type,
            'move': False,
            'application-id': self.main_window._id(),
            'ancestry': self.ancestry()[1],
            'structure-item': isinstance(self.item, QtStructureItem),
        }
        text = safe_decode(json.dumps(data))
        QtWidgets.QApplication.clipboard().setText(text)

    def paste(self):
        """Pastes clipboard data onto the current item, if possible."""
        data = self.clipboard_data()
        if data is None:
            return
        if data['type'] == 'item-existing':
            self.treeWidget().drop_event_item_existing(data,
                                                       target_treeitem=self)
        else:
            self.treeWidget().drop_event_item_new(data, target_treeitem=self)

    def clipboard_data(self):
        """Gets an item data dictionary from the clipboard.

        Returns
        -------
        dict, NoneType
            A data dictionary or None if no valid data was found.
        """
        import json
        from libqtopensesame.misc import drag_and_drop

        text = QtWidgets.QApplication.clipboard().text()
        try:
            data = json.loads(text)
        except Exception as e:
            return None
        if drag_and_drop.matches(data, ['item-snippet', 'item-existing']):
            return data
        return None
    
    def set_extra_info(self, extra_info=None):
        if extra_info is None:
            self.setText(1, '')
            return
        fixed_cond = self.item.syntax.fix_conditional_expression(extra_info)
        if fixed_cond != extra_info:
            self.setText(1, fixed_cond)
            return fixed_cond
        extra_info = extra_info.strip()
        extra_info_lower = extra_info.lower()
        color = None
        if extra_info_lower in ('always', '', 'true'):
            extra_info = 'True'
            color = QColor('green')
        elif extra_info_lower in ('never', 'false'):
            extra_info = 'False'
        self.setText(1, extra_info)
        if color is not None:
            self.setForeground(1, QBrush(color))
        else:
            self.setForeground(1, self.main_window.palette().text())
        return extra_info


# Alias for backwards compatibility
tree_item_item = TreeItemItem
